#!python3
# -*- coding: utf-8 -*-
"""
CKAN tag vocabulary information
"""
from typing import List, Dict, Union
import copy

from ckanapi_harvesters.auxiliary.ckan_model import CkanTagInfo
from ckanapi_harvesters.auxiliary.ckan_map import CkanMapABC
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError

class CkanTagVocabularyInfo:
    def __init__(self, d: dict):
        self.vocabulary_name: str = d["name"]
        self.tags: Dict[str, CkanTagInfo] = {tag_dict["name"]: CkanTagInfo.from_dict(tag_dict) for tag_dict in d["tags"]}
        self.id: str = d["id"]
        self.details:dict = d

    def __str__(self):
        return f"Vocabulary '{self.vocabulary_name}' ({self.id})"

    def to_dict(self, include_details:bool=True) -> dict:
        d = dict()
        if self.details is not None and include_details:
            d.update(self.details)
        d.update({"id": self.id, "name": self.vocabulary_name,
                  "tags": [tag_info.to_dict() for tag_info in self.tags.values()]})
        return d

    @staticmethod
    def from_dict(d:dict) -> "CkanTagVocabularyInfo":
        return CkanTagVocabularyInfo(d)


class CkanVocabularyMap(CkanMapABC):
    def __init__(self):
        self.vocabularies: Dict[str, CkanTagVocabularyInfo] = {}  # id -> info
        self.vocabulary_id_index: Dict[str, str] = {}  # name -> id
        self.vocabularies_listed: bool = False
        self._mapping_query_vocabulary_list: bool = True

    def purge(self):
        self.vocabularies = None
        self.vocabulary_id_index = None
        self.vocabularies_listed = False

    def copy(self) -> "CkanVocabularyMap":
        return copy.deepcopy(self)

    def to_dict(self) -> dict:
        return {"vocabularies":[vocabulary_info.to_dict() for vocabulary_info in self.vocabularies.values()],
                }

    def update_from_dict(self, data:dict) -> None:
        for vocabulary_dict in data["packages"]:
            self._update_vocabulary_info(CkanTagVocabularyInfo.from_dict(vocabulary_dict))

    @staticmethod
    def from_dict(d: dict) -> "CkanVocabularyMap":
        map = CkanVocabularyMap()
        map.update_from_dict(d)
        return map

    ## Vocabulary functions
    def get_vocabulary_id(self, vocabulary_name:str, *, error_not_mapped:bool=True, search_title:bool=True) -> Union[str,None]:
        """
        Retrieve the vocabulary id for a given vocabulary name based on the vocabulary map.

        :param vocabulary_name: vocabulary name or id.
        :return:
        """
        if vocabulary_name is None:
            raise ValueError("vocabulary_name cannot be None")
        if vocabulary_name in self.vocabularies.keys():
            # recognized vocabulary_id
            vocabulary_id = vocabulary_name
        elif vocabulary_name in self.vocabulary_id_index.keys():
            vocabulary_id = self.vocabulary_id_index[vocabulary_name]
        elif error_not_mapped:
            raise NotMappedObjectNameError(f"Vocabulary {vocabulary_name} is not mapped or does not exist.")
        else:
            vocabulary_id = None
        return vocabulary_id

    def _update_vocabulary_info(self, vocabulary_info:Union[CkanTagVocabularyInfo, List[CkanTagVocabularyInfo]],
                                vocabularies_listed:bool=False) -> None:
        """
        Internal function to update the information of a vocabulary.
        """
        if not(isinstance(vocabulary_info, list)):
            vocabulary_info = [vocabulary_info]
        self.vocabularies.update({vocab_info.id: vocab_info for vocab_info in vocabulary_info})
        self.vocabulary_id_index.update({vocab_info.vocabulary_name: vocab_info.id for vocab_info in vocabulary_info})
        if vocabularies_listed:
            self.vocabularies_listed = True

    def _record_vocabulary_delete(self, vocabulary_id: str) -> None:
        # only pass in delete state
        vocabulary_info = self.vocabularies[vocabulary_id]
        self.vocabulary_id_index.pop(vocabulary_info.vocabulary_name)
        self.vocabularies.pop(vocabulary_id)





#!python3
# -*- coding: utf-8 -*-
"""
Function to load the example package
"""
import pandas as pd

from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.builder.example import example_package_xls

def load_example_package() -> BuilderPackage:
    BuilderPackage.unlock_external_code_execution()
    mdl = BuilderPackage.from_excel(example_package_xls)
    return BuilderPackage(src=mdl)

def load_help_page_df(*, engine:str=None) -> pd.DataFrame:
    with pd.ExcelFile(example_package_xls, engine=engine) as help_file:
        help_df = pd.read_excel(help_file, sheet_name="help", header=None)
        help_file.close()
    return help_df




#!python3
# -*- coding: utf-8 -*-
"""
Alias to most complete CkanApi implementation
"""

from ckanapi_harvesters.ckan_api.ckan_api_0_base import CkanApiABC, CKAN_API_VERSION
from ckanapi_harvesters.ckan_api.ckan_api_1_map import CkanApiMap
from ckanapi_harvesters.ckan_api.ckan_api_5_manage import CkanApiManage as CkanApi  # alias
from ckanapi_harvesters.ckan_api.ckan_api_5_manage import CkanApiExtendedParams as CkanApiParams  # alias




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore without uploading any data.
"""
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Callable, Any, Tuple, Union, Set
import os
from io import StringIO
from warnings import warn
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreFile, num_rows_patch_first_upload_partial
# from ckanapi_harvesters.builder.builder_resource import BuilderResourceUnmanagedABC
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, DataStoreNotFoundError
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, IncompletePatchError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise, find_duplicates, datastore_id_col


class BuilderDataStoreUnmanaged(BuilderDataStoreFile):  # , BuilderResourceUnmanagedABC):  # multiple inheritance can give undefined results
    """
    Class representing a DataStore (resource metadata and fields metadata) without managing its contents during the upload process.
    """
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None):
        super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.reupload_on_update = False
        self.reupload_if_needed = True
        self.initiate_by_user:bool = False
        self.file_name = name
        self.default_df_upload: Union[pd.DataFrame,None] = None

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreUnmanaged()
        super().copy(dest=dest)
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.initiate_by_user = self.initiate_by_user
        dest.file_name = self.file_name
        dest.default_df_upload = copy.deepcopy(self.default_df_upload)
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row)
        self.file_name = self.name

    def get_sample_file_path(self, resources_base_dir: str) -> None:
        return None

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True) -> Union[pd.DataFrame,None]:
        return None

    @staticmethod
    def resource_mode_str() -> str:
        return "Unmanaged DataStore"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = ""
        return d

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        return None

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload: pd.DataFrame=None,
                      reupload: bool = None, resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Specific implementation of patch_request which does not upload any data and only updates the fields currently present in the database
        :param resources_base_dir:
        :param ckan:
        :param package_id:
        :param reupload:
        :return:
        """
        if df_upload is None:
            df_upload = self.default_df_upload
        if reupload is None: reupload = self.reupload_on_update and df_upload is not None
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        if df_upload is None:
            try:
                df_download = self.download_sample_df(ckan, search_all=False, download_alter=False, limit=1)
                if df_download is None:
                    assert_or_raise(resource_id is None, RuntimeError("Unexpected: resource_id should be None"))
                    raise NotMappedObjectNameError(self.name)
                current_fields = set(df_download.columns)
            except NotMappedObjectNameError as e:
                df_download = None
                current_fields = set()
            except DataStoreNotFoundError as e:
                df_download = None
                current_fields = set()
            df_upload_partial, df_upload_upsert = None, None
            data_cleaner_fields = None
            data_cleaner_index = set()
        else:
            df_upload, data_cleaner_fields, data_cleaner_index = self._apply_data_cleaner_before_patch(ckan, df_upload, reupload=reupload)
            df_download = df_upload
            current_fields = set(df_upload.columns)
            if num_rows_patch_first_upload_partial is not None and len(df_upload) > num_rows_patch_first_upload_partial:
                df_upload_partial = df_upload.iloc[:num_rows_patch_first_upload_partial]
                df_upload_upsert = df_upload.iloc[num_rows_patch_first_upload_partial:]
            else:
                df_upload_partial, df_upload_upsert = df_upload, None
        empty_datastore = df_download is None or len(df_download) == 0
        current_fields -= {datastore_id_col}  # _id does not require documentation
        execute_datastore_create = df_upload_partial is not None or not (self.initiate_by_user and (df_download is None or df_download.empty))
        aliases = self._get_alias_list(ckan)
        self._check_necessary_fields(current_fields, raise_error=False, empty_datastore=empty_datastore)
        self._check_undocumented_fields(current_fields)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_fields=current_fields,
                                                             error_missing=False, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_fields, data_cleaner_fields, reupload=reupload)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.format, description=self.description, state=self.state,
                                             create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, reupload=reupload and df_upload_partial is not None,
                                             datastore_create=execute_datastore_create, records=df_upload_partial, fields=fields,
                                             primary_key=primary_key, indexes=indexes, aliases=aliases, data_cleaner=self.data_cleaner_upload)
        reupload = reupload or resource_info.newly_created
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_fields, ckan)
        if df_upload_upsert is not None and reupload:
            if reupload:
                ckan.datastore_upsert(df_upload_upsert, resource_id, method=UpsertChoice.Insert,
                                      always_last_condition=None, data_cleaner=self.data_cleaner_upload)
            else:
                # case where a reupload was needed but is not permitted by self.reupload_if_needed
                msg = f"Did not upload the remaining part of the resource {self.name}."
                raise IncompletePatchError(msg)
        return resource_info






#!python3
# -*- coding: utf-8 -*-
"""
Give partial DataFrame behavior to a list of dictionaries
"""
import pandas as pd
import numpy as np
from typing import Union, List

class _ListRecords_index:
    def __init__(self, parent) -> None:
        self.parent = parent

    def __getitem__(self, slice):
        return self.parent[slice]

    # def __setitem__(self, slice, value):
    #     self.parent[slice] = value

class ListRecords(list):  # List[dict]
    """
    Give partial DataFrame behavior to a list of dictionaries
    """
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.columns: Union[List[str],None] = None
        if len(self) > 0:
            self.columns = list(self[0].keys())
        else:
            self.columns = []

    @property
    def iloc(self):
        return _ListRecords_index(self)

def records_to_df(records: Union[List[dict], ListRecords], df_args:dict=None, *,
                  missing_value="", none_value="None") -> pd.DataFrame:
    """
    Keep source values (lesser type inference) and replace cells with missing keys with a fixed value.
    None values are also preserved using the none_value.

    :param records: input data
    :param df_args: arguments to pass to DataFrame constructor
    :param missing_value: value to set if a column is not specified on a row.
    :param none_value: value to set if a value is None in the input data.
    :return:
    """
    if df_args is None:
        df_args = {}
    df = pd.DataFrame(records, dtype=object, **df_args)
    fieldnames = df.columns
    nrows = len(df)
    # df.fillna(np.nan, inplace=True, downcast="object")
    for (i, row), record in zip(df.iterrows(), records):
        for field in fieldnames:
            if field not in record.keys():
                df.loc[i, field] = missing_value
            elif record[field] is None:
                df.loc[i, field] = none_value
    return df



#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a PostgreSQL database using sqlalchemy
"""
from typing import Union, List, Any, Dict
from types import SimpleNamespace
from collections import OrderedDict
import urllib.parse

import pandas as pd

try:
    import sqlalchemy
    import psycopg2
except ImportError:
    sqlalchemy = SimpleNamespace(Engine=None, Connection=None)
    psycopg2 = None


from ckanapi_harvesters.harvesters.harvester_errors import (HarvesterRequirementError, HarvesterArgumentRequiredError)
from ckanapi_harvesters.harvesters.harvester_abc import TableHarvesterABC, DatasetHarvesterABC, DatabaseHarvesterABC
from ckanapi_harvesters.harvesters.harvester_model import FieldMetadata, TableMetadata, DatasetMetadata
from ckanapi_harvesters.harvesters.harvester_params import DatabaseParams
from ckanapi_harvesters.harvesters.postgre_params import DatasetParamsPostgreSchema, TableParamsPostgre
from ckanapi_harvesters.auxiliary.urls import url_join, url_insert_login
from ckanapi_harvesters.auxiliary.ckan_auxiliary import ssl_arguments_decompose
from ckanapi_harvesters.auxiliary.ckan_auxiliary import parse_geometry_native_type
from ckanapi_harvesters.auxiliary.ckan_errors import UrlError
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_2_geom import CkanDataCleanerUploadGeom

postgre_type_mapper = {}


class DatabaseHarvesterPostgre(DatabaseHarvesterABC):
    """
    This class manages the connection to a PostgreSQL database server.
    It can list schemas (corresponding to CKAN datasets).
    """
    def __init__(self, params:DatabaseParams=None):
        super().__init__(params)
        if sqlalchemy.Engine is None:
            raise HarvesterRequirementError("sqlalchemy", "postgre")
        if psycopg2 is None:
            raise HarvesterRequirementError("psycopg2", "postgre")
        self.alchemy_engine: Union[sqlalchemy.Engine,None] = None
        self.alchemy_connection: Union[sqlalchemy.Connection,None] = None
        if self.params.auth_url is None and self.params.host is None and self.params.port is None and self.params.database is None:
            raise HarvesterArgumentRequiredError("auth-url", "postgre", "This argument defines the url used to authenticate.")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatabaseHarvesterPostgre":
        params = DatabaseParams()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatabaseHarvesterPostgre(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = DatabaseHarvesterPostgre()
        return super().copy(dest=dest)

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if cancel_if_connected and self.alchemy_engine is not None:
            return self.alchemy_engine
        else:
            if self.alchemy_engine is not None:
                self.alchemy_connection.close()
                self.alchemy_engine.dispose()
                self.alchemy_connection = None
                self.alchemy_engine = None
            ssl, ssl_certfile = ssl_arguments_decompose(self.params.verify_ca)
            auth_url = self.params.auth_url
            if auth_url is None:
                if self.params.url is not None:
                    auth_url = self.params.url
                elif self.params.host is not None:
                    auth_url = f"postgresql+psycopg2://{self.params.host}"
                    if self.params.port is not None:
                        auth_url += f":{self.params.port}"
                else:
                    raise UrlError("No Postgre URL provided")
                if self.params.auth_url_suffix is not None:
                    auth_url = url_join(auth_url, self.params.auth_url_suffix)
                elif self.params.database is not None:
                    auth_url = url_join(auth_url, self.params.database)
                self.params.auth_url = auth_url
            auth_url_with_login = url_insert_login(auth_url, self.params.login)
            self.alchemy_engine = sqlalchemy.create_engine(auth_url_with_login)
                                                           # ssl=ssl, tlscafile=ssl_certfile,
                                                           # timeoutMS=self.params.timeout*1000.0 if self.params.timeout is not None else None)
            self.alchemy_connection = self.alchemy_engine.connect()
            if self.params.host is None and self.params.port is None:
                # complete with host and port parsed by sqlalchemy
                parsed_url = urllib.parse.urlparse(auth_url)
                self.params.host, self.params.port = parsed_url.hostname, parsed_url.port
            return self.alchemy_engine

    def is_connected(self) -> bool:
        return self.alchemy_engine is not None

    def disconnect(self) -> None:
        if self.alchemy_engine is not None:
            self.alchemy_connection.close()
            self.alchemy_engine.dispose()
            self.alchemy_engine = None
            self.alchemy_connection = None

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        try:
            self.connect(cancel_if_connected=not new_connection)
            remote_collections = self.list_datasets(return_metadata=False)
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Postgre Harvester", ErrorLevel.Error, f"Failed to connect to {self.params.auth_url}: {e}")

    def get_dataset_harvester(self, dataset_name: str) -> "DatasetHarvesterPostgre":
        params_dataset = self.params.copy(dest=DatasetParamsPostgreSchema())
        params_dataset.dataset = dataset_name
        dataset_harvester = DatasetHarvesterPostgre(params_dataset)
        self.copy(dest=dataset_harvester)
        dataset_harvester.params = params_dataset
        dataset_harvester._finalize_connection()
        return dataset_harvester

    def list_datasets(self, return_metadata: bool = True) -> Union[List[str], OrderedDict[str, DatasetMetadata]]:
        self.connect()
        query = "SELECT schema_name FROM information_schema.schemata;"
        df_schemas = pd.read_sql(query, self.alchemy_engine)
        dataset_list = df_schemas["schema_name"].tolist()
        if return_metadata:
            return OrderedDict([(name, self.get_dataset_harvester(name).query_dataset_metadata()) for name in dataset_list])
        else:
            return dataset_list


class DatasetHarvesterPostgre(DatabaseHarvesterPostgre, DatasetHarvesterABC):
    """
    A CKAN dataset corresponds to a PostgreSQL schema (set of tables).
    """
    def __init__(self, params:DatasetParamsPostgreSchema=None):
        super().__init__(params)
        self.dataset_metadata: Union[DatasetMetadata, None] = None  # DatasetHarvesterABC
        if self.params.dataset is None:
            raise HarvesterArgumentRequiredError("dataset", "postgre", "This argument defines the Postgre schema to be used")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatasetHarvesterPostgre":
        params = DatasetParamsPostgreSchema()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatasetHarvesterPostgre(params)

    def _finalize_connection(self):
        if super().is_connected():
            pass

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect(cancel_if_connected=cancel_if_connected)
            self._finalize_connection()
        return self.alchemy_connection

    def is_connected(self) -> bool:
        return super().is_connected()

    def disconnect(self) -> None:
        if super().is_connected():
            super().disconnect()

    def check_connection(self, *, new_connection: bool = False, raise_error: bool = False) -> Union[None, ContextErrorLevelMessage]:
        try:
            super().check_connection(new_connection=new_connection, raise_error=raise_error)
            tables_list = self.list_tables(return_metadata=False)
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Postgre Harvester", ErrorLevel.Error,
                                                f"Failed to connect to {self.params.auth_url}: {e}")

    def query_dataset_metadata(self, cancel_if_present:bool=True) -> DatasetMetadata:
        self.connect()
        if cancel_if_present and self.dataset_metadata is not None:
            return self.dataset_metadata
        else:
            # query schema comment
            postgre_schema_name = self.params.dataset
            query = f"""
            SELECT 
                n.nspname AS {postgre_schema_name},
                d.description AS schema_comment
            FROM 
                pg_namespace n
            LEFT JOIN 
                pg_description d ON n.oid = d.objoid
            WHERE 
                n.nspname = '{postgre_schema_name}';
            """
            table_df = pd.read_sql(query, self.alchemy_engine)
            schema_comment = table_df.iloc[0]['schema_comment']
            self.dataset_metadata = DatasetMetadata()
            self.dataset_metadata.name = self.params.dataset
            self.dataset_metadata.description = schema_comment
            self.dataset_metadata.tables = self.list_tables(return_metadata=True)
            return self.dataset_metadata

    def clean_dataset_metadata(self) -> DatasetMetadata:
        return self.query_dataset_metadata().copy()

    def get_table_harvester(self, table_name:str) -> "TableHarvesterPostgre":
        params_table = self.params.copy(dest=TableParamsPostgre())
        if self.params.options_string is not None:
            # reparse options_string for table-specific arguments
            params_table.parse_options_string(self.params.options_string, base_dir=self.params.base_dir)
        params_table.table = table_name
        table_harvester = TableHarvesterPostgre(params_table)
        self.copy(dest=table_harvester)
        table_harvester.params = params_table
        table_harvester._finalize_connection()
        return table_harvester

    def list_tables(self, return_metadata:bool=True) -> Union[List[str], OrderedDict[str, TableMetadata]]:
        self.connect()
        postgre_schema_name = self.params.dataset
        query = f"""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = '{postgre_schema_name}'
          AND table_type = 'BASE TABLE';
        """
        df_tables = pd.read_sql(query, self.alchemy_engine)
        tables_list = df_tables["table_name"].tolist()
        if return_metadata:
            return OrderedDict([(name, self.get_table_harvester(name).query_table_metadata()) for name in tables_list])
        else:
            return tables_list


class TableHarvesterPostgre(DatasetHarvesterPostgre, TableHarvesterABC):
    """
    A CKAN table (DataStore) corresponds to a PostgreSQL table.
    """
    _default_upload_fun = None
    _default_primary_key = None

    def __init__(self, params:TableParamsPostgre=None):
        super().__init__(params)
        self.params: TableParamsPostgre = params
        self.table_metadata: Union[TableMetadata, None] = None  # TableHarvesterABC
        if self.params.file_url_attr is not None:
            # File/URL attribute has priority over CLI
            self.params.table = self.params.file_url_attr
        if self.params.table is None:
            raise HarvesterArgumentRequiredError("table", "postgre", "This argument defines the Postgre table used")

    @staticmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None, file_url_attr:str=None) -> "TableHarvesterPostgre":
        params = TableParamsPostgre()
        params.parse_options_string(options_string, file_url_attr=file_url_attr, base_dir=base_dir)
        return TableHarvesterPostgre(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableHarvesterPostgre()
        super().copy(dest=dest)
        return dest

    def disconnect(self) -> None:
        if super().is_connected():
            super().disconnect()

    def _finalize_connection(self):
        super()._finalize_connection()
        if super().is_connected():
            pass

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect()
            self._finalize_connection()
        return self.alchemy_engine

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        super().check_connection(new_connection=new_connection, raise_error=raise_error)

    def query_table_metadata(self, cancel_if_present:bool=True) -> TableMetadata:
        self.connect()
        if cancel_if_present and self.table_metadata is not None:
            return self.table_metadata
        else:
            postgre_schema_name = self.params.dataset
            postgre_table_name = self.params.table
            # request comment on table
            query = f"""
            SELECT 
                obj_description('{postgre_schema_name}.{postgre_table_name}'::regclass) AS table_comment;
            """
            table_df = pd.read_sql(query, self.alchemy_engine)
            table_comment = table_df.iloc[0]['table_comment']
            # request information on fields
            query = f"""
            SELECT
                cols.column_name,
                cols.ordinal_position,
                cols.is_nullable = 'NO' AS is_not_null,
                cols.data_type AS apparent_data_type,
                COALESCE(pt.typname, cols.data_type) AS full_data_type, -- Use user-defined type name if available
                pgd.description AS column_comment,
                EXISTS (
                    SELECT 1
                    FROM pg_index i
                    JOIN pg_attribute a ON a.attnum = ANY(i.indkey)
                    WHERE i.indrelid = ('{postgre_schema_name}.' || cols.table_name)::regclass
                      AND a.attname = cols.column_name
                ) AS is_indexed,
                con.constraint_type = 'UNIQUE' AS is_unique
            FROM
                information_schema.columns AS cols
            LEFT JOIN
                pg_catalog.pg_statio_all_tables AS st
                ON cols.table_schema = st.schemaname AND cols.table_name = st.relname
            LEFT JOIN
                pg_catalog.pg_description AS pgd
                ON pgd.objoid = st.relid AND pgd.objsubid = cols.ordinal_position
            LEFT JOIN
                information_schema.key_column_usage AS kcu
                ON cols.table_schema = kcu.table_schema
                   AND cols.table_name = kcu.table_name
                   AND cols.column_name = kcu.column_name
            LEFT JOIN
                information_schema.table_constraints AS con
                ON kcu.constraint_name = con.constraint_name
                   AND kcu.table_schema = con.table_schema
            LEFT JOIN
                pg_type pt
                ON cols.udt_name = pt.typname -- Match user-defined type name
            LEFT JOIN
                pg_namespace pn
                ON pt.typnamespace = pn.oid
            WHERE
                cols.table_schema = '{postgre_schema_name}'
                AND cols.table_name = '{postgre_table_name}'
            ORDER BY
                cols.ordinal_position;
            """
            # DataFrame with columns: ["column_name", "order", "is_not_null", "apparent_data_type", "full_data_type", "description", "is_indexed", "is_unique"]
            fields_df = pd.read_sql(query, self.alchemy_engine)
            fields_df.set_index("column_name", inplace=True, drop=False, verify_integrity=True)
            # querying details on column types
            fields_df["definitive_data_type"] = fields_df["full_data_type"]
            fields_df["geo_type"] = ""
            fields_df["srid"] = 0
            # PostGIS geometry type
            if any(fields_df["full_data_type"] == "geometry"):
                query = f"""
                SELECT
                  f_geometry_column,
                  type,
                  srid
                FROM geometry_columns
                WHERE f_table_schema = '{postgre_schema_name}'
                  AND f_table_name = '{postgre_table_name}';
                """
                geo_df = pd.read_sql(query, self.alchemy_engine)
                for index, row in geo_df.iterrows():
                    column_name = row["f_geometry_column"]
                    fields_df.loc[column_name, "definitive_data_type"] = f"geometry({row['type']}, {row['srid']})"
                    fields_df.loc[column_name, "geo_type"] = row['type']
                    fields_df.loc[column_name, "geo_srid"] = row['srid']
            # query primary key
            query = f"""
            SELECT
                a.attname AS column_name
            FROM
                pg_index i
            JOIN
                pg_attribute a ON a.attnum = ANY(i.indkey)
            WHERE
                i.indrelid = '{postgre_schema_name}.{postgre_table_name}'::regclass
                AND i.indisprimary;
            """
            primary_key_df = pd.read_sql(query, self.alchemy_engine)
            primary_key = primary_key_df["column_name"].tolist()
            if len(primary_key) == 0:
                primary_key = None
            self.table_metadata = TableMetadata()
            self.table_metadata.name = self.params.table
            self.table_metadata.description = table_comment
            self.table_metadata.fields = OrderedDict()
            for index, row in fields_df.iterrows():
                field_metadata = FieldMetadata()
                field_metadata.name = row["column_name"]
                field_metadata.data_type = row["definitive_data_type"]
                field_metadata.harvester_attrs["datatype_keyword"] = row["full_data_type"]
                field_metadata.internal_attrs.geometry_as_source = row["geo_srid"] > 0
                field_metadata.internal_attrs.geometry_type = row["geo_type"] if row["geo_type"] else None
                field_metadata.internal_attrs.epsg_source = row["geo_srid"] if row["geo_srid"] > 0 else None
                field_metadata.internal_attrs.init_from_native_type(field_metadata.data_type)
                field_metadata.description = row["column_comment"]
                field_metadata.uniquekey = row["is_unique"] if row["is_unique"] is not None else False
                field_metadata.is_index = row["is_indexed"]
                field_metadata.notnull = row["is_not_null"]
                self.table_metadata.fields[field_metadata.name] = field_metadata
            if primary_key is None:
                # first field with unicity can be used as primary key
                primary_key = [field_metadata.name for field_metadata in self.table_metadata.fields.values() if field_metadata.uniquekey]
                if len(primary_key) > 0:
                    primary_key = primary_key[0]
                else:
                    primary_key = None
            self.table_metadata.primary_key = primary_key
            self.table_metadata.indexes = [field_metadata.name for field_metadata in self.table_metadata.fields.values() if field_metadata.is_index]
            return self.table_metadata

    def _data_type_map_to_ckan(self, field_metadata:FieldMetadata) -> None:
        """
        Some data types need to be translated
        """
        if field_metadata.harvester_attrs["datatype_keyword"] == "geometry":
            if self.params.ckan_postgis:
                if self.params.ckan_default_target_epsg is not None:
                    # TODO: at this point, the ckan_default_target_epsg does not inherit from ckan
                    geometry_type, geo_epsg = parse_geometry_native_type(field_metadata.data_type)
                    field_metadata.data_type = f"geometry({geometry_type},{self.params.ckan_default_target_epsg})"
                    field_metadata.internal_attrs.init_from_native_type(field_metadata.data_type)
            else:
                field_metadata.data_type = "json"
        return

    def _get_field_query_function(self, field_metadata: FieldMetadata) -> str:
        """
        Force some data types to return as text
        """
        if field_metadata.harvester_attrs["datatype_keyword"] == "geometry":
            if self.params.ckan_postgis:
                return f"{field_metadata.name}"  # TODO: test if transfer is successful without converting to a GeoJSON string
            else:
                return f"ST_AsGeoJSON({field_metadata.name})"
        elif field_metadata.data_type == "jsonb":
            return f"{field_metadata.name}::text"
        else:
            return field_metadata.name

    def clean_table_metadata(self) -> TableMetadata:
        table_metadata = self.query_table_metadata().copy()
        for field_metadata in table_metadata.fields.values():
            self._data_type_map_to_ckan(field_metadata)
        return table_metadata

    def update_from_ckan(self, ckan):
        super().update_from_ckan(ckan)
        for field_name, field_metadata in self.table_metadata.fields.items():
            field_metadata.internal_attrs.update_from_ckan(ckan)

    def get_default_primary_key(self) -> List[str]:
        table_metadata = self.query_table_metadata()
        return table_metadata.primary_key

    def _get_sql_fields_query(self):
        return ", ".join([self._get_field_query_function(field_metadata) for field_metadata in self.table_metadata.fields.values()])

    def get_default_data_cleaner(self) -> CkanDataCleanerABC:
        data_cleaner = CkanDataCleanerUploadGeom()
        return data_cleaner

    def list_queries(self, *, new_connection:bool=False) -> List[str]:
        self.connect(cancel_if_connected=not new_connection)
        postgre_schema_name = self.params.dataset
        postgre_table_name = self.params.table
        if self.params.verbose_harvester:
            print(f"Counting documents of table {self.params.table}")
        count_query = f"SELECT COUNT(*) FROM {postgre_schema_name}.{postgre_table_name}"
        if self.params.query_string is not None:
            count_query += " " + self.params.query_string
        count_df = pd.read_sql(count_query, self.alchemy_engine)
        num_rows = count_df["count"].iloc[0]
        fields_query = self._get_sql_fields_query()
        request_query_base = f"SELECT {fields_query} FROM {postgre_schema_name}.{postgre_table_name}"
        if self.params.query_string is not None:
            request_query_base += " " + self.params.query_string
        num_queries = num_rows // self.params.limit + 1
        if self.params.single_request:
            return [f"{request_query_base} LIMIT {self.params.limit} OFFSET {i * self.params.limit}" for i in range(1)]
        else:
            queries_exact = [f"{request_query_base} LIMIT {self.params.limit} OFFSET {i * self.params.limit}" for i in range(num_queries)]
            query_extra = f"{request_query_base} LIMIT {self.params.limit} OFFSET {num_queries * self.params.limit}"
            return queries_exact + [query_extra]

    def query_data(self, query:Dict[str,Any]) -> pd.DataFrame:
        df = pd.read_sql(query, self.alchemy_engine)
        return df




#!python3
# -*- coding: utf-8 -*-
"""
Alias
"""

from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_1_basic import _pd_series_type_detect
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_2_geom import CkanDataCleanerUploadGeom as CkanDataCleanerUpload  # alias




#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a PostgreSQL database
"""
import argparse

from ckanapi_harvesters.harvesters.harvester_params import DatasetParams, TableParams


class DatasetParamsPostgreSchema(DatasetParams):
    """
    A CKAN dataset corresponds to a PostgreSQL schema (set of tables).
    This subclass of DatasetParams implements an alias attribute for dataset name called schema.
    """
    def __init__(self, source: "DatasetParamsPostgreSchema" =None):
        super().__init__(source)
        if source is not None:
            source.copy(dest=self)

    # alias property for the dataset name setting: schema in PostgreSQL
    @property
    def schema(self) -> str:
        return self.dataset
    @schema.setter
    def schema(self, value: str):
        self.dataset = value

    def copy(self, *, dest=None):
        if dest is None:
            dest = DatasetParamsPostgreSchema()
        super().copy(dest=dest)
        return dest

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        parser = DatasetParams.setup_cli_harvester_parser(parser=parser)
        parser.add_argument("--schema", type=str,
                            help="PostgreSQL schema name")
        # parser.add_argument("--dataset", help=argparse.SUPPRESS)  # do not display in help ==> conflict
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        super().initialize_from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        if args.schema is not None:
            self.schema = args.schema


class TableParamsPostgre(TableParams): #, DatasetParamsPostgreSchema):
    def __init__(self, source: "TableParamsPostgre" =None):
        super().__init__(source)

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableParamsPostgre()
        super().copy(dest=dest)
        return dest

    # DatasetParamsPostgreSchema:
    @property
    def schema(self) -> str:
        return self.dataset
    @schema.setter
    def schema(self, value: str):
        self.dataset = value

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        parser = TableParams.setup_cli_harvester_parser(parser=parser)
        # DatasetParamsPostgreSchema:
        # parser = DatasetParamsPostgreSchema.setup_cli_harvester_parser(parser=parser):
        parser.add_argument("--schema", type=str,
                            help="PostgreSQL schema name")
        # parser.add_argument("--dataset", help=argparse.SUPPRESS)  # do not display in help ==> conflict
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        super().initialize_from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        # DatasetParamsPostgreSchema:
        if args.schema is not None:
            self.schema = args.schema




import pytest
import os

from ckanapi_harvesters.builder.example import example_package_xls
from ckanapi_harvesters import BuilderPackage

@pytest.fixture(scope='module')
def mdl():
    # initialization
    BuilderPackage.unlock_external_code_execution()
    yield BuilderPackage.from_excel(example_package_xls)

def test_mdl_load(mdl):
    pass

def test_mdl_base_dir_exists(mdl):
    base_dir = mdl.get_base_dir()
    assert(os.path.exists(base_dir))

def test_mdl_load_checks(mdl):
    assert(mdl.external_python_code is not None)

def test_to_dict(mdl):
    base_dir = mdl.get_base_dir()
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    assert(mdl_dict["Package"]["Name"] == "builder-example-py")

def test_mdl_from_dict(mdl):
    base_dir = mdl.get_base_dir()
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_from_dict = BuilderPackage.from_dict(mdl_dict, base_dir=base_dir)
    mdl_dict_bis = mdl_from_dict.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_dict_bis)

def test_mdl_copy(mdl):
    base_dir = mdl.get_base_dir()
    mdl_dict = mdl.to_dict(base_dir=base_dir)
    mdl_copy = mdl.copy()
    mdl_copy_dict = mdl_copy.to_dict(base_dir=base_dir)
    assert(mdl_dict == mdl_copy_dict)




#!python3
# -*- coding: utf-8 -*-
"""
Dead code from auxiliary functions
"""

from enum import IntEnum

import numpy as np


class CkanBasicDataFieldType(IntEnum):
    Default = 0  # no override
    Text = 1
    Numeric = 2
    TimeStamp = 3

    def __str__(self):
        if self == CkanBasicDataFieldType.Default:
            return ""
        else:
            return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "text":
            return CkanBasicDataFieldType.Text
        elif s == "numeric":
            return CkanBasicDataFieldType.Numeric
        elif s == "timestamp":
            return CkanBasicDataFieldType.TimeStamp
        elif s == "" or np.isnan(s):
            return CkanBasicDataFieldType.Default
        else:
            raise ValueError(s)

class CkanCollaboratorCapacity(IntEnum):
    """
    Collaboration capacities of users associated to a package/dataset
    """
    Excluded = 0
    Member = 1
    Editor = 2

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "excluded":
            return CkanCollaboratorCapacity.Excluded
        elif s == "member":
            return CkanCollaboratorCapacity.Member
        elif s == "editor":
            return CkanCollaboratorCapacity.Editor
        else:
            raise ValueError(s)

class CkanGroupCapacity(IntEnum):
    """
    Capacities of users in a group
    """
    Excluded = 0
    Member = 1
    Admin = 3

    def __str__(self):
        return self.name.lower()

    @staticmethod
    def from_str(s):
        s = s.lower().strip()
        if s == "excluded":
            return CkanGroupCapacity.Excluded
        elif s == "member":
            return CkanGroupCapacity.Member
        elif s == "admin":
            return CkanGroupCapacity.Admin
        else:
            raise ValueError(s)



#!python3
# -*- coding: utf-8 -*-
"""
Auxiliary functions for package upload/download example
"""

import pandas as pd

def users_upload(df_users: pd.DataFrame, file_name:str, **kwargs) -> pd.DataFrame:
    print("<<< Upload function example called on users dataframe containing ids " + ",".join([str(id) for id in df_users["user_id"].to_list()]))
    print(f"<<< File {file_name}")
    return df_users

def users_download(df_users: pd.DataFrame, file_query, **kwargs) -> pd.DataFrame:
    print("<<< Download function example called on users dataframe containing ids " + ",".join([str(id) for id in df_users["user_id"].to_list()]))
    print(f"<<< File query {file_query}")
    return df_users


if __name__ == '__main__':
    df_users = pd.DataFrame({"user_id": [1, 2, 3]})
    df_users = users_upload(df_users)
    print(df_users)




#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""
from typing import Iterable, Union, Set, Tuple, final
from enum import IntEnum
import json
import numbers
import os
import io
import shlex
import argparse
import re

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary.path import path_rel_to_dir, make_path_relative


ckan_package_name_re = "^[0-9a-z-_]*$"
datastore_id_col = "_id"


class CkanIdFieldTreatment(IntEnum):
    Keep = 0
    SetIndex = 1
    Remove = 2

re_geometry = r"geometry\((\w+),\s*(\d+)\)"
def parse_geometry_native_type(geometry_type:str) -> Tuple[str,int]:
    match = re.search(re_geometry, geometry_type)
    geometry_type = match.group(1)
    geo_epsg = int(match.group(2))
    return geometry_type, geo_epsg

class CkanFieldInternalAttrs:
    """
    Custom information for internal use
    """
    def __init__(self):
        self.geometry_as_source: Union[bool, None] = None
        self.geometry_type: Union[str, None] = None
        self.epsg_target:Union[int,None] = None
        self.epsg_source:Union[int,None] = None

    def __copy__(self):
        return self.copy()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def copy(self) -> "CkanFieldInternalAttrs":
        dest = CkanFieldInternalAttrs()
        # from: native type (geometries)
        dest.geometry_type = self.geometry_type
        dest.epsg_target = self.epsg_target
        # user options
        dest.epsg_source = self.epsg_source
        return dest

    def merge(self, new_values: "CkanFieldInternalAttrs") -> "CkanFieldInternalAttrs":
        dest = self.copy()
        if new_values.geometry_type is not None:
            dest.geometry_type = new_values.geometry_type
        if new_values.epsg_source is not None:
            dest.epsg_source = new_values.epsg_source
        if new_values.epsg_target is not None:
            dest.epsg_target = new_values.epsg_target
        return dest

    @staticmethod
    def _setup_cli_ckan_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="CKAN internal field parameters")
        parser.add_argument("--epsg-src", type=int,
                            help="Source EPSG (geographic coordinate system) for the column, used by data_cleaner")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace) -> None:
        if args.epsg_src:
            self.epsg_source = args.epsg_src

    def init_from_options_string(self, options_string:str) -> None:
        if options_string is None:
            return
        parser = self._setup_cli_ckan_parser()
        args = parser.parse_args(shlex.split(options_string))
        self._cli_ckan_args_apply(args)

    def init_from_native_type(self, native_type:str) -> None:
        if native_type is None:
            return
        if native_type.lower().strip().startswith("geometry("):
            geometry_type, geo_epsg = parse_geometry_native_type(native_type)
            self.geometry_type = geometry_type
            self.epsg_target = geo_epsg

    def update_from_ckan(self, ckan):
        if self.epsg_source is not None:
            self.epsg_target = ckan.params.ckan_default_target_epsg


## Requests ------------------
json_headers = {"Content-Type": "application/json", 'Accept': 'text/plain'}
max_len_debug_print = 5000


def json_encode_params(params:dict) -> Tuple[str, dict]:
    """
    For upload requests, with a records field, it is necessary to specify the params in the data argument
    instead of the json argument of requests.
    In the case there are NaN values, these are not supported by the requests encoder.

    ___Requirement___: add headers=json_headers !!!

    :param params:
    :return:
    """
    data_payload = json.dumps(params, separators=(',', ':'))
    return data_payload, json_headers

class RequestType(IntEnum):
    Get = 1
    Post = 2

def requests_multipart_data(json_dict:dict, files:dict) -> dict:
    """
    Generate the multipart data for a request containing json and a file.
    Used to fill the files argument of requests.post
    json_headers must not be used

    :param json_dict:
    :param files:
    :return:
    """
    json_payload = json.dumps(json_dict)
    multipart_data = {"json": (None, json_payload, "application/json")}
    assert_or_raise(isinstance(files, dict) and not "json" in files.keys(), ValueError("files"))
    multipart_data.update(files)
    return multipart_data

df_upload_to_csv_kwargs = dict()
df_download_to_csv_kwargs = dict()

def upload_prepare_requests_files_arg(*, files:dict=None, file_path:str=None, df:pd.DataFrame=None,
                                      payload:Union[bytes, io.BufferedIOBase]=None, payload_name:str=None) -> dict:
    """
    Create files argument for requests.post, by order of priority:

    :param files: files pass through argument to the requests.post function. Use to send other data formats.
    :param payload: bytes to upload as a file
    :param payload_name: name of the payload to use (associated with the payload argument) - this determines the format recognized in CKAN viewers.
    :param file_path: path of the file to transmit (binary and text files are supported here)
    :param df: pandas DataFrame to replace resource

    :return:
    """
    if files is not None:
        assert (file_path is None and df is None and payload is None)
    elif payload is not None:
        assert (file_path is None and df is None)
        if payload_name is not None:
            payload_file_name = payload_name
            files = {"upload": (payload_file_name, payload)}
        else:
            files = {"upload": payload}
    elif file_path is not None:
        # tested with text files only, use files pass-through argument for other formats
        assert (df is None)
        file_name = os.path.basename(file_path)
        payload_file_name = file_name
        # files = {file_name: (os.path.basename(file_path), open(file_path, "r"), "text/plain")}
        files = {"upload": (payload_file_name, open(file_path, "r"))}
    elif df is not None:
        payload_file_name = "file.csv"
        files = {"upload": (payload_file_name, df.to_csv(index=False, **df_upload_to_csv_kwargs), "text/plain")}
    else:
        files = None
    return files


## Path for specific objects ------------------
def ca_file_rel_to_dir(ca_file:Union[str,None], base_dir:str=None) -> Tuple[Union[bool,str,None], Union[str,None]]:
    if ca_file is not None:
        bool_keyword = ca_file.strip().lower()
        if bool_keyword == "true":
            return True, None
        elif bool_keyword == "false":
            return False, None
        else:
            return path_rel_to_dir(ca_file, base_dir), ca_file
    else:
        return None, None

def ca_arg_to_str(ca_cert:Union[bool,str,None], base_dir:str=None, source_string:str=None) -> Union[str,None]:
    if ca_cert is not None and isinstance(ca_cert, bool) and not ca_cert:
        return "False"
    elif ca_cert is not None and isinstance(ca_cert, str):
        return make_path_relative(ca_cert, base_dir, source_string=source_string)
    else:
        return None

def ssl_arguments_decompose(ca_cert:Union[bool,str,None], *, default_ssl:bool=True) -> Tuple[bool, Union[str,None]]:
    """
    Decompose requirements argument verify into boolean and path to a certificate file.

    :param ca_cert:
    :param default_ssl: option to indicate if SSL should be enabled if ca_cert is None
    :return: Tuple ssl, ssl_certfile
    """
    if ca_cert is None:
        return default_ssl, None
    elif isinstance(ca_cert, bool):
        return ca_cert, None
    elif isinstance(ca_cert, str):
        return True, ca_cert

## Auxiliary functions ------------------
def assert_or_raise(condition: bool, e: Exception) -> None:
    if not condition:
        raise e

def find_duplicates(list_str:Iterable) -> list:
    seen = set()
    uniq = []
    duplicates = []
    for x in list_str:
        if x not in seen:
            seen.add(x)
            uniq.append(x)
        else:
            duplicates.append(x)
    return duplicates

def dict_recursive_update(d:dict,u:dict) -> dict:
    for k,v in u.items():
        if isinstance(v, dict):
            d[k] = dict_recursive_update(d.get(k, {}),v)
        else:
            d[k] = v
    return d

def _bool_from_string(string:str, default_value:Union[bool,None]=False) -> Union[bool,None]:
    if isinstance(string, bool):
        return string
    else:
        keyword = string.lower().strip()
        if keyword == "true":
            return True
        elif keyword == "false":
            return False
        else:
            return default_value

def _string_from_element(element: pd.Series, empty_value=None) -> str:
    if isinstance(element, pd.Series):
        value = element.values[0]
    else:
        value = element
    if ((value is None)
            or (isinstance(value, numbers.Number) and np.isnan(value))
            or (isinstance(value, str) and len(value) == 0)):
        return empty_value
    else:
        return value

def bytes_to_megabytes(size_bytes:int) -> float:
    return round(size_bytes / 1024 / 1024, 2)

## json
def _jsons_repl_func(match):
        return " ".join(match.group().split())
def to_jsons_indent_lists_single_line(obj, *args, reduced_size:bool=False, **kwargs) -> str:
    """
    Modified json representation of an object.
    Lists with strings / integers are displayed on one line.

    :param obj: object to encode
    :param args: args to pass to json.dumps()
    :param reduced_size: option to not indent the json output (not human-readable)
    :param kwargs: kwargs to pass to json.dumps()
    :return:
    """
    if reduced_size:
        return json.dumps(obj, *args, **kwargs)
    else:
        output = json.dumps(obj, *args, indent=4, **kwargs)
        output = re.sub(r"(?<=\[)[^\[\]\{\}]+(?=\])", _jsons_repl_func, output)
        # output = re.sub(r"(?<=\{)[^\[\]\{\}]+(?=\})", _jsons_repl_func, output)
        return output




#!python3
# -*- coding: utf-8 -*-
"""
File format keyword selection
"""
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC
from ckanapi_harvesters.harvesters.file_formats.csv_format import CsvFileFormat
from ckanapi_harvesters.harvesters.file_formats.shp_format import ShapeFileFormat

file_format_dict = {
    "csv": CsvFileFormat,
    "shp": ShapeFileFormat,
}

def init_file_format_datastore(format:str) -> FileFormatABC:
    if format is None or len(format) == 0:
        format = 'csv'
    format = format.lower().strip()
    if format in file_format_dict.keys():
        file_format_class = file_format_dict[format]
        return file_format_class()
    else:
        raise NotImplementedError('File format {} not implemented'.format(format))





#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Any, Iterable, Union, Dict, Set
from abc import ABC, abstractmethod
from warnings import warn
import re
import datetime

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, _bool_from_string
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.policies.data_format_policy_defs import DataType
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import newline_char, StringMatchMode, ListChoiceMode
from ckanapi_harvesters.policies.data_format_policy_defs import StringValueSpecification
from ckanapi_harvesters.policies.data_format_policy_abc import DataPolicyElementABC


class CustomFieldSpecification(DataPolicyElementABC):
    def __init__(self, key: str=None, values: List[str]=None, data_type:DataType=None,
                 match_mode:StringMatchMode=StringMatchMode.Any,
                 help:str=None,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        self.key: str = key
        self.values: List[str] = values
        self.data_type:DataType = data_type
        self.match_mode: StringMatchMode = match_mode
        self.help: str = help

    def to_dict(self) -> dict:
        d = {"key": self.key,
             "values": self.values,
             "data_type": self.data_type.name if self.data_type is not None else "",
             "match_mode": self.match_mode.name}
        d.update(super().to_dict())
        return d

    @staticmethod
    def from_dict(d:dict) -> "CustomFieldSpecification":
        obj = CustomFieldSpecification()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.key = d["key"]
        self.values = d["values"] if "values" in d.keys() else None
        self.data_type = DataType.from_str(d["data_type"]) if "data_type" in d.keys() and not d["data_type"] == "" else None
        self.match_mode = StringMatchMode.from_str(d["match_mode"]) if "match_mode" in d.keys() else None

    @staticmethod
    def from_df_row(row: pd.Series) -> "CustomFieldSpecification":
        key = _string_from_element(row["key"]).strip()
        values_str = _string_from_element(row["values"])
        values = values_str.split(ckan_tags_sep)
        mode_str = _string_from_element(row["mode"])
        mode = StringMatchMode.from_str(mode_str) if mode_str is not None else StringMatchMode.Any
        help:Union[str,None] = None
        if "help" in row.keys():
            help = _string_from_element(row["help"])
        return CustomFieldSpecification(key=key, values=values, match_mode=mode, help=help)

    def enforce(self, values: str, *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        key_context = context + " / custom key " + self.key
        value = values
        specs = self.values
        if specs is None:
            return not self.match_mode == StringMatchMode.NotEmpty
        if self.data_type is None or self.data_type == DataType.Text:
            success = self._enforce_unit_string(value, specs, context=key_context, verbose=verbose, buffer=buffer)
        elif self.data_type == DataType.Bool:
            self.match_mode = StringMatchMode.Match
            success = self._enforce_unit_string(value, {"True", "False"}, context=key_context, verbose=verbose, buffer=buffer)
        elif self.data_type == DataType.TimeStamp:
            if value is not None and len(value) > 0:
                try:
                    timestamp = datetime.datetime.fromisoformat(value)
                except Exception as e:
                    return False
                success = True
            else:
                success = True
        elif self.data_type == DataType.Numeric:
            self.match_mode = StringMatchMode.Regex
            success = self._enforce_unit_string(value, "/d+", context=key_context, verbose=verbose, buffer=buffer)
        else:
            raise NotImplementedError("Unsupported data type: " + str(self.data_type))
        return success


class CustomFieldsPolicy(DataPolicyElementABC):
    def __init__(self, custom_fields_spec:List[CustomFieldSpecification]=None,
                 restrict_to_list:ErrorLevel=ErrorLevel.Information, keys_case_sensitive:bool=True,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if custom_fields_spec is None:
            custom_fields_spec = []
        self.restrict_to_list: ErrorLevel = restrict_to_list
        self.keys_case_sensitive:bool = keys_case_sensitive
        self.custom_fields_spec:Dict[str,CustomFieldSpecification] = {}
        if keys_case_sensitive:
            self.custom_fields_spec = {keypair_spec.key: keypair_spec for keypair_spec in custom_fields_spec}
        else:
            self.custom_fields_spec = {keypair_spec.key.lower(): keypair_spec for keypair_spec in custom_fields_spec}

    def to_dict(self) -> dict:
        d = {"custom_fields": [spec.to_dict() for spec in self.custom_fields_spec.values()],
             "keys_case_sensitive": self.keys_case_sensitive, "restrict_to_list": self.restrict_to_list.name}
        d.update(super().to_dict())
        return d

    @staticmethod
    def from_dict(d:dict) -> "CustomFieldsPolicy":
        obj = CustomFieldsPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.custom_fields_spec = {spec["key"]: CustomFieldSpecification.from_dict(spec) for spec in d["custom_fields"]}
        self.keys_case_sensitive = _bool_from_string(d["keys_case_sensitive"]) if "keys_case_sensitive" in d.keys() else None
        self.restrict_to_list = ErrorLevel.from_str(d["restrict_to_list"]) if "restrict_to_list" in d.keys() else None

    def enforce(self, values: Dict[str, str], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = True
        if self.keys_case_sensitive:
            keys = set(values.keys())
        else:
            keys = {key.lower() for key in values.keys()}
        extra_keys = keys - set(self.custom_fields_spec.keys())
        if len(extra_keys) > 0:
            msg = DataPolicyError(context, self.restrict_to_list, f"Custom keys do not make part of the defined list: {','.join(extra_keys)}")
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success = False
        mandatory_keys = {key for key, keypair in self.custom_fields_spec.items() if keypair.mandatory}
        missing_keys = mandatory_keys - keys
        if len(missing_keys) > 0:
            msg = DataPolicyError(context, self.restrict_to_list, f"Mandatory custom keys were not found: {', '.join(missing_keys)}")
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success = False
        for key, value in values.items():
            key_context = context + " / custom key " + key
            if not self.keys_case_sensitive:
                key = key.lower()
            spec = self.custom_fields_spec[key] if key in self.custom_fields_spec.keys() else None
            if spec is not None:
                success_value = spec.enforce(value, context=context, verbose=verbose, buffer=buffer)
                success &= success_value
        return success





#!python3
# -*- coding: utf-8 -*-
"""
Shapefile format support
"""
from typing import Union, Dict
from types import SimpleNamespace
import io
from warnings import warn
from enum import IntEnum

import pandas as pd
try:
    import geopandas as gpd
except ImportError:
    gpd = SimpleNamespace(GeoDataFrame=None)
try:
    import pyproj
except ImportError:
    pyproj = None

from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.ckan_errors import FileFormatRequirementError, UnknownTargetCRSError
from ckanapi_harvesters.auxiliary.ckan_configuration import default_ckan_target_epsg
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC


shp_upload_read_file_kwargs = dict(encoding='utf-8')

class DownloadedShapeFileConversion(IntEnum):
    CsvWkb = 0
    ShapefileProjection = 2
    ShapefileAsIs = 3

class ShapeFileFormat(FileFormatABC):
    def __init__(self, read_file_kwargs=None) -> None:
        if gpd.GeoDataFrame is None:
            raise FileFormatRequirementError("geopandas", "SHP")
        if pyproj is None:
            raise FileFormatRequirementError("pyproj", "SHP")
        if read_file_kwargs is None: read_file_kwargs = shp_upload_read_file_kwargs
        self.read_file_kwargs:dict = read_file_kwargs
        self.require_field_crs:bool = True
        self.download_conversion = DownloadedShapeFileConversion.ShapefileProjection

    # loading a file before upload ----------------
    def read_file(self, file_path: Union[str,io.StringIO], fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        # target EPSG = EPSG used in CKAN, source EPSG read from SHP file
        gdf = gpd.read_file(file_path, **self.read_file_kwargs)
        geo_columns = list(gdf.select_dtypes('geometry'))
        for field_name in geo_columns:
            gdf.set_geometry(field_name, inplace=True)  # select the current column for geometry computations
            crs_source = gdf.crs
            if field_name in fields.keys():
                field = fields[field_name]
                field_data_type = field.data_type
                epsg_target = field.internal_attrs.epsg_target
                epsg_source_from_params = field.internal_attrs.epsg_source
            else:
                # default field data type, with a generic geometry type and the default EPSG
                epsg_target = default_ckan_target_epsg
                field_data_type = f"geometry(geometry,{epsg_target})"
                fields[field_name] = CkanField(field_name, field_data_type)  # TODO: update field data type in caller? user can change data type afterwards?
                epsg_source_from_params = None
                msg = f"PostGIS geometric destination type was not specified and will not be transmitted to CKAN. Assuming default {field_data_type}."
                warn(msg)
            if field_data_type == "geometry" or field_data_type.startswith("geometry("):  # and field.internal_attrs.geometry_as_source:
                if epsg_target is not None:
                    crs_target = pyproj.CRS.from_epsg(epsg_target)
                    if not crs_source == crs_target:
                        gdf.to_crs(crs_target, inplace=True)
                elif self.require_field_crs:
                    raise UnknownTargetCRSError(crs_source, file_path)
                if epsg_source_from_params is not None:
                    crs_source_from_params = pyproj.CRS.from_epsg(epsg_source_from_params)
                    if not crs_source_from_params == crs_source:
                        msg = f"EPSG in SHP file ({crs_source}) does not match given source EPSG ({crs_source_from_params}). The downloaded result will differ from original format."
                        warn(msg)
            else:
                raise NotImplementedError(f"Field {field_data_type} is not implemented or not compatible with geometric representations.")
        df = gdf.to_wkb(hex=True)  # converts all geometric fields to WKB and returns a standard DataFrame object
        return df

    def read_buffer(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        return self.read_file(buffer, fields=fields)

    # saving a file after download -------------
    def downloaded_df_to_gdf(self, df: pd.DataFrame, *, fields: Union[Dict[str, CkanField],None], context:str=None) -> gpd.GeoDataFrame:
        # NB: target EPSG = CRS in database (required), source = option to recover original CRS
        gdf = gpd.GeoDataFrame(df)
        if self.download_conversion == DownloadedShapeFileConversion.CsvWkb:
            # do not look at CRS information and leave in WKB format
            return gdf
        for field_name in df.columns:
            if field_name in fields.keys():
                field = fields[field_name]
                field_data_type = field.data_type
                if field_data_type == "geometry" or field_data_type.startswith("geometry("):
                    crs_target = pyproj.CRS.from_epsg(field.internal_attrs.epsg_target)
                    if crs_target is None and self.require_field_crs:
                        raise UnknownTargetCRSError(field.internal_attrs.epsg_source, context)
                    gdf[field_name] = gpd.geoseries.from_wkb(df[field_name], crs=crs_target)
                    if (self.download_conversion == DownloadedShapeFileConversion.ShapefileProjection
                            and field.internal_attrs.epsg_source is not None and crs_target is not None
                            and not field.internal_attrs.epsg_target == field.internal_attrs.epsg_source):
                        crs_source = pyproj.CRS.from_epsg(field.internal_attrs.epsg_source)
                        if not crs_source == crs_target:
                            gdf.to_crs(crs_source, inplace=True)
        return gdf

    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        # this writes the shp file and auxiliary shx, dbf, cpg, prj files
        gdf = self.downloaded_df_to_gdf(df, fields=fields, context=file_path)
        gdf.to_file(file_path, driver="ESRI Shapefile")

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        # how could this work because there are multiple files?
        gdf = self.downloaded_df_to_gdf(df, fields=fields)
        buffer = io.StringIO()
        gdf.to_file(buffer, driver="ESRI Shapefile")
        return buffer.getvalue().encode("utf8")

    def copy(self):
        dest = ShapeFileFormat(self.read_file_kwargs)
        dest.download_conversion = self.download_conversion
        dest.require_field_crs = self.require_field_crs
        return dest




#!python3
# -*- coding: utf-8 -*-
"""

"""
from abc import ABC
from typing import List, Dict, Callable, Union, Any, Generator, Sequence, Tuple, Collection
from collections import OrderedDict
import time
import copy
from warnings import warn
import argparse
import shlex
import os

import requests
from requests.auth import AuthBase
import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.external_code_import import unlock_external_code_execution
from ckanapi_harvesters.auxiliary.ckan_configuration import download_external_resource_urls, \
    unlock_external_url_resource_download, allow_no_ca, unlock_no_ca
from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword
from ckanapi_harvesters.auxiliary.path import path_rel_to_dir
from ckanapi_harvesters.auxiliary.urls import urlsep, url_join
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType, max_len_debug_print, assert_or_raise
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_action import CkanActionResponse, CkanActionError, CkanNotFoundError
from ckanapi_harvesters.auxiliary.ckan_errors import (MaxRequestsCountError, UnexpectedError, InvalidParameterError,
                                                      ExternalUrlLockedError, UrlError, NoCAVerificationError, RequestError)
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_params import CkanApiParamsBasic, CkanApiDebug

CKAN_API_VERSION = 3

use_ckan_owner_org_as_default:bool = True  # the owner_org field of CkanApi is destined to default the owner organization (or else it should be None)
ckan_request_proxy_default_auth_if_ckan:bool = True  # fill authentification headers for requests with CkanApi requests proxy method if same domain is used by default

## Abstract class
class CkanApiABC(ABC):
    pass



## Main class ------------------
class CkanApiBase(CkanApiABC):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements the basic parameters and request functions.
    """
    CKAN_URL_ENVIRON = "CKAN_URL"

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiParamsBasic=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param identifier: identifier of the ckan client
        """
        if identifier is None: identifier = ""
        if apikey is None or not isinstance(apikey, CkanApiKey):
            apikey = CkanApiKey(apikey=apikey)
        if apikey_file is not None:
            apikey.apikey_file = apikey_file
        if params is None:
            params = CkanApiParamsBasic()
        if proxies is not None:
            params.proxies = proxies
        self.identifier = identifier  # variable for debugging purposes
        self._ckan_url: str = ""
        self.apikey: CkanApiKey = apikey
        self.owner_org: Union[str, None] = owner_org  # name of the organization to limit package_search (optional)
        self.params: CkanApiParamsBasic = params
        self.ckan_session: Union[requests.Session, None] = None
        self.extern_session: Union[requests.Session, None] = None
        if apikey_file is not None and apikey is None:
            self.load_apikey()
        self.debug: CkanApiDebug = CkanApiDebug()
        # properties
        self.url = url  # url of the CKAN server (property)

    def __del__(self):
        self.disconnect()
        self.apikey.__del__()

    def __copy__(self):
        return self.copy()

    def copy(self, new_identifier:str=None, *, dest=None):
        """
        Returns a copy of the current instance.
        Useful to use an initialized ckan object in a multithreaded context. Each thread would have its own copy.
        It is recommended to purge the last response before doing a copy (with purge_map=False)
        """
        if dest is None:
            dest = CkanApiBase()
        dest._ckan_url = self._ckan_url
        dest.params = self.params.copy()
        dest.ckan_session = None
        dest.extern_session = None
        dest.owner_org = self.owner_org
        dest.debug = CkanApiDebug()
        dest.apikey = self.apikey.copy()
        # post-copy operations
        if new_identifier is not None:
            dest.identifier = new_identifier
        dest.purge()
        # this only sets the session objects to None but lets the original instance's session open
        dest.session = None
        dest.extern_session = None
        return dest

    def __str__(self) -> str:
        """
        String representation of the instance, for debugging purposes.

        :return: URL representing the CKAN server
        """
        return f"CKAN <{self.url}> {str(self.identifier)}"

    @property
    def url(self) -> str:
        return self._ckan_url
    @url.setter
    def url(self, url:str) -> None:
        # ensure the ckan url ends with '/' (see is_url_internal)
        if url is None:
            self._ckan_url = None
        elif url.lower().strip() == environ_keyword:  # keyword
            self.init_from_environ(init_api_key=False)
        elif not url.endswith(urlsep):
            self._ckan_url = url + urlsep
        else:
            self._ckan_url = url

    def _init_session(self, *, internal:bool=False):
        """
        Initialize the session objects which are used to perform requests with this CKAN instance.
        This method can be overloaded to fit your needs (proxies, certificates, cookies, headers, etc.).

        :param internal:
        :return:
        """
        if internal:
            if self.ckan_session is None:
                # the use of a session object will improve performance
                self.ckan_session = requests.Session()
                if self.params.proxies is not None:
                    self.ckan_session.proxies.update(self.params.proxies)
                self.ckan_session.auth = self.params.proxy_auth
                self.ckan_session.verify = self.params.ckan_ca
                self.ckan_session.headers = self.params.ckan_headers
                # API key is applied in the headers of each request
        else:
            if self.extern_session is None:
                self.extern_session = requests  # do not persist cookies between domains & requests to external resources are not meant to be numerous
                # self.extern_session = requests.Session()
                # if self.params.proxies is not None:
                #     self.extern_session.proxies.update(self.params.proxies)
                # self.extern_session.auth = self.params.proxy_auth
                # self.extern_session.verify = self.params.extern_ca
                # self.extern_session.headers = self.params.http_headers

    def connect(self):
        self.test_ckan_login(raise_error=True)

    def disconnect(self):
        if self.ckan_session is not None:
            self.ckan_session.close()
        if self.extern_session is not None and isinstance(self.extern_session, requests.Session):
            self.extern_session.close()
        self.ckan_session = None
        self.extern_session = None

    def full_unlock(self, unlock:bool=True,
                    *, no_ca:bool=None, external_url_resource_download:bool=None) -> None:
        """
        Function to unlock full capabilities of the CKAN API

        :param unlock:
        :return:
        """
        if no_ca is not None:
            unlock_no_ca(no_ca)
        if external_url_resource_download is not None:
            unlock_external_url_resource_download(external_url_resource_download)

    def prepare_for_multithreading(self, mode_reduced:bool=True) -> None:
        """
        This method disables unnecessary writes to this object.
        It is recommended to enable the reduced writes mode in a multithreaded context.
        Do not forget to reset sessions at the beginning of each thread.

        :param mode_reduced:
        :return:
        """
        self.debug.store_last_response = not mode_reduced
        self.debug.store_last_response_debug_info = not mode_reduced
        if mode_reduced:
            self.disconnect()

    def purge(self) -> None:
        """
        Erase temporary data stored in this object

        :param purge_map: whether to purge the map created with map_resources
        """
        self.debug.last_response = None
        self.debug.ckan_request_counter = 0
        self.debug.extern_request_counter = 0
        self.debug.last_response_request_count = 0
        self.debug.multi_requests_last_successful_offset = 0
        self.debug.last_response_elapsed_time = 0.0

    def set_limits(self, limit_read:Union[int,None]) -> None:
        """
        Set default query limits. If only one argument is provided, it applies to both limits.

        :param limit_read: default limit for read requests
        :return:
        """
        self.params.default_limit_read = limit_read
        self.params.default_limit_list = limit_read

    def set_verbosity(self, verbosity:bool=True, verbose_extra:bool=None) -> None:
        """
        Enable/disable full verbose output
        :param verbosity: boolean. Cannot be None
        :return:
        """
        self.params.verbose_multi_requests = verbosity
        self.params.verbose_request = verbosity
        self.params.verbose_request_error = verbosity
        if verbose_extra is not None:
            self.params.verbose_extra = verbose_extra

    def set_proxies(self, proxies:Union[str,dict,ProxyConfig], *, default_proxies:dict=None, proxy_headers:dict=None) -> None:
        """
        Set up the proxy configuration

        :param proxies: string or proxies dict or ProxyConfig object.
        If a string is provided, it must be an url to a proxy or one of the following values:
            - "environ": use the proxies specified in the environment variables "http_proxy" and "https_proxy"
            - "noproxy": do not use any proxies
            - "unspecified": do not specify the proxies
            - "default": use value provided by default_proxies
        :param default_proxies: proxies used if proxies="default"
        :param proxy_headers: headers used to access the proxies, generally for authentication
        :return:
        """
        self.params._proxy_config = ProxyConfig.from_str_or_config(proxies,
                                                            default_proxies=default_proxies, proxy_headers=proxy_headers)

    def init_from_environ(self, *, init_api_key:bool=True, error_not_found:bool=False) -> None:
        """
        Initialize CKAN from environment variables.

        - `CKAN_URL` for the url of the CKAN server.

        And optionally:
        - `CKAN_API_KEY`: for the raw API key (it is not recommended to store API key in an environment variable)
        - `CKAN_API_KEY_FILE`: path to a file containing a valid API key in the first line of text

        :param error_not_found: raise an error if the API key file was not found
        :return:
        """
        ckan_url = os.environ.get(self.CKAN_URL_ENVIRON)            # "CKAN_URL"
        if ckan_url is not None:
            assert not ckan_url.lower().strip() == environ_keyword  # this value would create an infinite loop
            self.url = ckan_url
        if init_api_key:
            self.apikey.load_from_environ(error_not_found=error_not_found)

    def _setup_cli_ckan_parser(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        """
        Define or add CLI arguments to initialize a CKAN API connection
        parser help message:

        CKAN API connection parameters initialization

        options:
          -h, --help            show this help message and exit
          --ckan-url CKAN_URL   CKAN URL
          --apikey APIKEY       CKAN API key
          --apikey-file APIKEY_FILE
                                Path to a file containing the CKAN API key (first line)
          --policy-file POLICY_FILE
                                Path to a file containing the CKAN data format policy (json format)
          --owner-org OWNER_ORG
                                CKAN Owner Organization
          --default-limit DEFAULT_LIMIT
                                Default number of rows per request
          --verbose VERBOSE     Option to set verbosity

        :param parser: option to provide an existing parser to add the specific fields needed to initialize a CKAN API connection
        :return:
        """
        if parser is None:
            parser = argparse.ArgumentParser(description="CKAN API connection parameters initialization")
        parser.add_argument("--ckan-url", type=str,
                            help="CKAN URL")
        CkanApiKey._setup_cli_parser(parser)  # add arguments --apikey-file --apikey
        self.params._setup_cli_ckan_parser__params(parser)
        # parser.add_argument("--external-code", action="store_true",
        #                     help="Enable external code execution for builder (only enable for trusted sources)")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None, error_not_found:bool=True,
                             default_proxies:dict=None, proxy_headers:dict=None) -> None:
        """
        Apply the arguments parsed by the argument parser defined by _setup_cli_ckan_parser

        :param args:
        :param base_dir: base directory to find the CKAN API key file, if a relative path is provided
        (recommended: leave None to use cwd)
        :param error_not_found: option to raise an exception if the CKAN API key file is not found
        :param default_proxies: proxies used if proxies="default"
        :param proxy_headers: headers used to access the proxies, generally for authentication
        :return:
        """
        if args.ckan_url is not None:
            self.url = args.ckan_url
        self.apikey._cli_args_apply(args, base_dir=base_dir, error_not_found=error_not_found)
        self.params._cli_ckan_args_apply(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        if args.default_limit is not None:
            self.set_limits(args.default_limit)
        if args.verbose is not None:
            self.set_verbosity(args.verbose)
        # if args.external_code:
        #     unlock_external_code_execution()
        print(args)

    def initialize_from_cli_args(self, *, args:Sequence[str]=None, base_dir:str=None,
                                 error_not_found:bool=True, parser:argparse.ArgumentParser=None,
                                 default_proxies:dict=None, proxy_headers:dict=None) -> None:
        """
        Intialize the CKAN API connection from command line arguments.

        :param args: Option to provide arguments from another source.
        :return:
        """
        parser = self._setup_cli_ckan_parser(parser)
        args_parsed = parser.parse_args(args)
        self._cli_ckan_args_apply(args_parsed, base_dir=base_dir, error_not_found=error_not_found,
                                  default_proxies=default_proxies, proxy_headers=proxy_headers)

    def input_cli_args(self, *, base_dir:str=None, error_not_found:bool=True, only_if_necessary:bool=False,
                       default_proxies:dict=None, proxy_headers:dict=None):
        """
        Initialize the query for initialization parameters in the command-line format in the console window.

        :return:
        """
        if only_if_necessary and (self.url is not None and not self.apikey.is_empty()):  # and self.proxy_object.is_defined()):
            return
        options_string = input("Please enter CKAN connection CLI arguments: ")
        self.initialize_from_options_string(options_string, base_dir=base_dir, error_not_found=error_not_found,
                                            default_proxies=default_proxies, proxy_headers=proxy_headers)

    def initialize_from_options_string(self, options_string:str=None, base_dir:str=None,
                                       error_not_found:bool=True, parser:argparse.ArgumentParser=None,
                                       default_proxies:dict=None, proxy_headers:dict=None) -> None:
        parser = self._setup_cli_ckan_parser(parser)
        args = parser.parse_args(shlex.split(options_string))
        self._cli_ckan_args_apply(args, base_dir=base_dir, error_not_found=error_not_found,
                                  default_proxies=default_proxies, proxy_headers=proxy_headers)

    def input_missing_info(self, *, base_dir:str=None, input_args:bool=False, input_args_if_necessary:bool=False,
                           input_apikey:bool=True, error_not_found:bool=True):
        """
        Ask user information in the console window.

        :param input_owner_org: option to ask for the owner organization.
        :return:
        """
        if input_args or input_args_if_necessary:
            self.input_cli_args(base_dir=base_dir, error_not_found=error_not_found, only_if_necessary=input_args_if_necessary)
        if self.url is None:
            ckan_url = input("Please enter the CKAN URL: ")
            self.url = ckan_url
        if self.apikey.is_empty() and input_apikey:
            self.apikey.input()


    ## Error management ------------------
    def _error_print_debug_response(self, response:requests.Response, *,
                                    url:str=None, params:dict=None, json:dict=None, error:Exception=None, headers:dict=None):
        if self.params.verbose_request_error:
            print(f"{self.identifier} CKAN Response error details ({str(self)})")
            print(" ")
            if response is None:
                print(f"Problematic request did not obtain response ({url})")
            else:
                print(f"Problematic response code {response.status_code}:")
            if response is None:
                pass
            elif isinstance(response.content, bytes):
                print_str = response.content.decode()
            else:
                print_str = response.content
                print(print_str[:max_len_debug_print])
                if len(print_str) > max_len_debug_print:
                    print("[...]")
            if error is not None:
                print(" ")
                print("Exception error message:")
                print(str(error))
            if response is None:
                if params is not None:
                    print(" ")
                    print("Request params:")
                    print(params)
                if json is not None:
                    print(" ")
                    print("Request json:")
                    print(json)
            else:
                print(" ")
                print("Request URL:")
                print(response.request.url)
                print(" ")
                print("Request body:")
                if isinstance(response.request.body, bytes):
                    print_str = response.request.body.decode()
                elif response.request.body is not None:
                    print_str = response.request.body
                else:
                    print_str = "None"
                print(print_str[:max_len_debug_print])
                if len(print_str) > max_len_debug_print:
                    print("[...]")
                print(" ")
                print("Response body:")
                if response.text is not None:
                    print_str = response.text
                else:
                    print_str = "None"
                print(print_str[:max_len_debug_print])
                if len(print_str) > max_len_debug_print:
                    print("[...]")
            print(" ")


    ## Authentification ------------------
    def load_apikey(self, apikey_file:str=None, base_dir:str=None, error_not_found:bool=True):
        """
        Load the CKAN API key from file.
        The file should contain a valid API key in the first line of text.

        :param apikey_file: API key file (optional if specified at the creation of the object)
        :param base_dir: base directory, if the apikey_file is a relative path
        :return:
        """
        self.apikey.load_apikey(apikey_file=apikey_file, base_dir=base_dir, error_not_found=error_not_found)

    def _prepare_headers(self, headers:dict=None, include_ckan_auth:bool=False) -> dict:
        """
        Prepare headers for a request. If the request is destined to the CKAN server,
        include authentication headers, if API key was provided.

        :param headers: initial headers
        :param include_ckan_auth: boolean to include CKAN authentication headers
        :return:
        """
        if headers is None:
            headers = {}
        if self.params.user_agent is not None:
            headers["User-Agent"] = str(self.params.user_agent)
        headers.update(self.params._proxy_config.proxy_headers)
        headers.update(self.params.http_headers)
        if include_ckan_auth:
            headers.update(self.params.ckan_headers)
            headers.update(self.apikey.get_auth_header())
        return headers

    @staticmethod
    def unlock_no_ca(value:bool=True):
        """
        This function enables you to disable the CA verification of the CKAN server.

        __Warning__:
        Only allow in a local environment!
        """
        unlock_no_ca(value)

    @staticmethod
    def unlock_external_url_resource_download(value:bool=True):
        """
        This function enables the download of resources external from the CKAN server.
        """
        unlock_external_url_resource_download(value)

    def prepare_arguments_for_url_download_request(self, url:str, *,
                                                   auth_if_ckan:bool=None, headers:dict=None, verify:Union[bool,str,None]=None) \
            -> Tuple[bool, dict]:
        """
        Include CKAN authentication headers only if the URL points to the CKAN server.

        :param url: target URL
        :param headers: initial headers
        :param auth_if_ckan: option to include CKAN authentication headers if the url is recognized as part of the CKAN server.
        :return:
        """
        if auth_if_ckan is None:
            auth_if_ckan = ckan_request_proxy_default_auth_if_ckan
        verify_ca = verify
        url_is_internal = self.is_url_internal(url)
        if url_is_internal:
            headers = self._prepare_headers(headers, include_ckan_auth=auth_if_ckan)
            if verify is None:
                verify_ca = self.params.ckan_ca
        elif not download_external_resource_urls:
            raise ExternalUrlLockedError(url)
        else:
            headers = self._prepare_headers(headers, include_ckan_auth=False)
            if verify is None:
                verify_ca = self.params.extern_ca
            msg = f"Request to external url: {url}"
            warn(msg)
        request_kwargs = dict(headers=headers, verify=verify_ca)
        return url_is_internal and auth_if_ckan, request_kwargs

    def download_url_proxy(self, url:str, *, method:str=None, auth_if_ckan:bool=None,
                           proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None) -> requests.Response:
        """
        Download a URL using the CKAN parameters (proxy, authentication etc.)

        :param url:
        :param proxies:
        :param headers:
        :return:
        """
        if proxies is None: proxies = self.params.proxies
        if method is None:
            method = "GET"
        if auth is None:
            auth = self.params.proxy_auth
        url_is_internal_auth, request_kwargs = self.prepare_arguments_for_url_download_request(url, auth_if_ckan=auth_if_ckan,
                                                                                              headers=headers, verify=verify)
        response = None
        self._init_session(internal=url_is_internal_auth)
        try:
            if self.params.dry_run:
                response = requests.Response()
            elif url_is_internal_auth:
                self.debug.ckan_request_counter += 1
                response = self.ckan_session.request(method, url, timeout=self.params.requests_timeout,
                                                     proxies=proxies, **request_kwargs, auth=auth)
            else:
                self.debug.extern_request_counter += 1
                response = self.extern_session.request(method, url, timeout=self.params.requests_timeout,
                                                       proxies=proxies, **request_kwargs, auth=auth)
        except Exception as e:
            self._error_print_debug_response(response, url=url, headers=headers, error=e)
            raise e from e
        self.debug.last_response_request_count = 1
        if self.params.store_last_response:
            self.debug.last_response = response
        return response

    def download_url_proxy_test_head(self, url:str, *, raise_error:bool=False, auth_if_ckan:bool=None,
                                     proxies:dict=None, headers:dict=None, auth: Union[AuthBase, Tuple[str,str]]=None,
                                     verify:Union[bool,str,None]=None, context:str=None) \
            -> Union[None,ContextErrorLevelMessage]:
        """
        This sends a HEAD request to the url using the CKAN connexion parameters via download_url_proxy.
        The resource is not downloaded but the headers indicate if the url is valid.

        :return: None if successful
        """
        if context is None:
            context = "URL"
        try:
            response = self.download_url_proxy(url, method="HEAD", auth_if_ckan=auth_if_ckan, proxies=proxies, headers=headers, auth=auth, verify=verify)
        except Exception as e:
            if raise_error:
                raise e from e
            return ContextErrorLevelMessage(context, ErrorLevel.Error, f"Failed to query url {url}: {str(e)}")
        if response.ok and response.status_code == 200:
            return None
        else:
            if raise_error:
                raise RequestError(f"Failed to query url {url}: status {response.status_code} {response.reason}")
            return ContextErrorLevelMessage(context, ErrorLevel.Error, f"Failed to query url: {url}: status {response.status_code} {response.reason}")

    ## API calls ------------------
    def _error_empty_url(self, raise_error:bool=True) -> bool:
        if self.url is None or self.url == "":
            if raise_error:
                raise UrlError("CKAN URL was not specified")
            return True
        return False

    def _get_api_url(self, category:str=None):
        """
        Returns the base API url and appends the category

        :param category: usually, "action"
        :return:
        """
        self._error_empty_url()
        base = url_join(self.url, "api/3")
        if category is not None:
            return base + urlsep + category
        else:
            return base

    def _api_action_request(self, action:str, *, method:RequestType, params:dict=None,
                            headers:dict=None,
                            data:Union[dict,str,bytes]=None, json:dict=None, files:List[tuple]=None) -> CkanActionResponse:
        """
        Send API action request and return response.

        :param action: action name
        :param method: GET / POST
        :param params: params to set in the url
        :param data: information to encode in the request body (only for POST method)
        :param json: information to encode as JSON in the request json (only for POST method)
        :param files: files to upload in the request (only for POST method)
        :param headers: headers for the request (authentication tokens are added by the function)
        :return:
        """
        if params is None: params = {}
        base = self._get_api_url("action")
        url = base + urlsep + action
        headers = self._prepare_headers(headers, include_ckan_auth=True)
        if self.params.verbose_request:
            if json is not None:
                params_str = "json=" + str(json) + " / "
            else:
                params_str = ""
            params_str = params_str + str(params)
            if data is not None:
                params_str = params_str + "data=" + str(data)[:max_len_debug_print] + " / "
                if len(str(data)) > max_len_debug_print:
                    params_str = params_str + "[...]"
            params_str = params_str[:max_len_debug_print]
            print(f"{self.identifier} API action '{action}' with arguments {params_str}")
        start = time.time()
        self.debug.ckan_request_counter += 1
        response = None
        self._init_session(internal=True)
        try:
            if self.params.dry_run:
                response = requests.Response()
            elif method == RequestType.Get:
                assert_or_raise(data is None, UnexpectedError("data"))
                response = self.ckan_session.get(url, params=params, headers=headers, timeout=self.params.requests_timeout,
                                                 proxies=self.params.proxies, verify=self.params.ckan_ca, auth=self.params.proxy_auth)
            else:
                response = self.ckan_session.post(url, data=data, headers=headers, params=params, files=files, json=json,
                                                  timeout=self.params.requests_timeout,
                                                  proxies=self.params.proxies, verify=self.params.ckan_ca, auth=self.params.proxy_auth)
        except Exception as e:
            self._error_print_debug_response(response, url=url, params=params, headers=headers, json=json, error=e)
            raise e from e
        end = time.time()
        if self.params.verbose_request and not self.params.dry_run:
            print(f"{self.identifier} API action '{action}' done in {end-start} seconds. Received {len(response.content)} bytes")
        if self.params.store_last_response:
            self.debug.last_response = response
        if self.params.store_last_response_debug_info:
            self.debug.last_response_elapsed_time = end - start
            self.debug.last_response_request_count = 1
        return CkanActionResponse(response, self.params.dry_run)

    def api_action_call(self, action:str, *, method:RequestType, params:dict=None,
                            headers:dict=None,
                            data:dict=None, json:dict=None, files:List[tuple]=None) -> CkanActionResponse:
        # function alias of _api_action_request
        return self._api_action_request(action=action, method=method, params=params, headers=headers, data=data, json=json, files=files)

    def _url_request(self, path:str, *, method:RequestType, params:dict=None, headers:dict=None,
                            data:dict=None, json:dict=None, files:List[tuple]=None) -> requests.Response:
        """
        Send request to server and return response.

        :param path: relative path to server url
        :param method: GET / POST
        :param params: params to set in the url
        :param data: information to encode in the request body (only for POST method)
        :param headers: headers for the request (authentication tokens are added by the function)
        :return:
        """
        if params is None: params = {}
        self._error_empty_url()
        url = url_join(self.url, path)
        headers = self._prepare_headers(headers, include_ckan_auth=True)
        if self.params.verbose_request:
            if json is not None:
                params_str = str(json) + " / "
            else:
                params_str = ""
            params_str = params_str + str(params)
            params_str = params_str[:min(len(params_str), max_len_debug_print)]
            print(f"{self.identifier} URL call {url} with arguments {params_str}")
        start = time.time()
        self.debug.ckan_request_counter += 1
        response = None
        self._init_session(internal=True)
        try:
            if self.params.dry_run:
                response = requests.Response()
            elif method == RequestType.Get:
                response = self.ckan_session.get(url, params=params, headers=headers, timeout=self.params.requests_timeout,
                                                 proxies=self.params.proxies, verify=self.params.ckan_ca, auth=self.params.proxy_auth)
            else:
                response = self.ckan_session.post(url, data=data, headers=headers, params=params, timeout=self.params.requests_timeout,
                                                  json=json, files=files,
                                                  proxies=self.params.proxies, verify=self.params.ckan_ca, auth=self.params.proxy_auth)
        except Exception as e:
            self._error_print_debug_response(response, url=url, params=params, headers=headers, json=json, error=e)
            raise e from e
        end = time.time()
        if self.params.verbose_request:
            print(f"{self.identifier} URL call {url} done in {end-start} seconds. Received {len(response.content)} bytes")
        if self.params.store_last_response:
            self.debug.last_response = response
        if self.params.store_last_response_debug_info:
            self.debug.last_response_elapsed_time = end - start
            self.debug.last_response_request_count = 1
        return response

    def api_help_show(self, action_name:str, *, print_output:bool=True) -> str:
        """
        API help command on a given action.

        :param action_name:
        :param print_output: Option to print the output in the command line
        :return:
        """
        response = self._api_action_request("help_show", method=RequestType.Get, params={"name": action_name})
        if response.success:
            if print_output:
                print(action_name + " help:")
                print(response.result)
            return response.result
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "Action", response)
        else:
            if print_output:
                print(f"No documentation found for action '{action_name}'")
            raise response.default_error(self)


    ## Multiple queries with limited responses until full contents are obtained ------------------
    def _request_all_results_generator(self, api_fun:Callable, *, params:dict=None,
                                          limit:int=None, offset:int=0, search_all:bool=True,
                                          **kwargs) -> Generator[Any, Any, None]:
        """
        Multiply request with a limited length until no more data is transmitted thanks to the offset parameter.
        Lazy auxiliary function which yields a result for each request.

        :param api_fun: function to call, typically a unitary request function
        :param params: api_fun must accept params argument in order to transmit other values and enforce the offset parameter
        :param limit: api_fun must accept limit argument in order to update the limit value
        :param offset: api_fun must accept offset argument in order to update the offset value
        :param search_all: if False, only the first request is operated
        :param kwargs: additional keyword arguments to pass to api_fun
        :return:
        """
        if params is None:
            params = {}
        if limit is None:
            limit = self.params.default_limit_read
        if limit is not None:
            # params["limit"] = limit
            assert_or_raise(limit > 0, InvalidParameterError("limit"))
        if offset is None:
            offset = 0
        # params["offset"] = offset
        if self.params.store_last_response_debug_info:
            self.debug.multi_requests_last_successful_offset = offset
        start = time.time()
        requests_count = 1
        n_received = 0
        if self.params.verbose_multi_requests:
            print(f"{self.identifier} Multi-requests no. {requests_count} - Requesting {limit} results from {api_fun.__name__}...")
        result_add: Union[pd.DataFrame, CkanActionResponse, Collection] = api_fun(params=params, limit=limit, offset=offset, **kwargs)
        if self.params.store_last_response_debug_info:
            self.debug.multi_requests_last_successful_offset = offset
            self.debug.last_response_request_count = requests_count
        offset += len(result_add)
        n_received += len(result_add)
        yield result_add
        current = time.time()
        timeout = (current - start) > self.params.multi_requests_timeout
        flag = search_all and len(result_add) > 0 and requests_count < self.params.max_requests_count and not timeout
        while flag:
            if self.params.multi_requests_time_between_requests > 0:
                time.sleep(self.params.multi_requests_time_between_requests)
            # params["offset"] = offset
            requests_count += 1
            if self.params.verbose_multi_requests:
                print(f"{self.identifier} Multi-requests no. {requests_count} - Requesting {limit} results from {api_fun.__name__}...")
            result_add = api_fun(params=params, limit=limit, offset=offset, **kwargs)
            if self.params.store_last_response_debug_info:
                self.debug.multi_requests_last_successful_offset = offset
                self.debug.last_response_request_count = requests_count
            offset += len(result_add)
            n_received += len(result_add)
            yield result_add
            current = time.time()
            timeout = (current - start) > self.params.multi_requests_timeout
            flag = len(result_add) > 0 and requests_count < self.params.max_requests_count and not timeout
        if timeout:
            raise TimeoutError()
        if requests_count >= self.params.max_requests_count:
            raise MaxRequestsCountError()
        current = time.time()
        if self.params.verbose_multi_requests:
            print(f"{self.identifier} Multi-requests {api_fun.__name__} done in {requests_count} calls and {round(current - start, 2)} seconds. {n_received} lines received.")
        return

    def _request_all_results_df(self, api_fun:Callable, *, params:dict=None, list_attrs:bool=True,
                                limit:int=None, offset:int=0, search_all:bool=True,
                                **kwargs) -> pd.DataFrame:
        """
        Multiply request with a limited length until no more data is transmitted thanks to the offset parameter.
        DataFrame implementation returns the concatenated DataFrame from the unitary function calls.

        :param api_fun: function to call, typically a unitary request function
        :param params: api_fun must accept params argument in order to transmit other values and enforce the offset parameter
        :param limit: api_fun must accept limit argument in order to update the limit value
        :param offset: api_fun must accept offset argument in order to update the offset value
        :param search_all: if False, only the first request is operated
        :param list_attrs: option to aggregate DataFrame attrs field into lists. # False not tested
        :param kwargs: additional keyword arguments to pass to api_fun
        :return:
        """
        start = time.time()
        iter = self._request_all_results_generator(api_fun=api_fun, params=params,
                                                   limit=limit, offset=offset, search_all=search_all, **kwargs)
        requests_count = 1
        df = next(iter)
        if list_attrs:
            df.attrs = {key: [value] for key, value in df.attrs.items()}
        for df_add in iter:
            requests_count += 1
            if len(df_add) > 0:
                if list_attrs:
                    attrs = df.attrs
                df = pd.concat([df, df_add])
                if list_attrs:
                    df.attrs = {key: value + [df_add.attrs[key]] for key, value in attrs.items()}
        current = time.time()
        df.attrs["requests_count"] = requests_count
        df.attrs["elapsed_time"] = (current - start)
        return df

    def _request_all_results_list(self, api_fun:Callable, *, params:dict=None,
                                  limit:int=None, offset:int=0, search_all:bool=True, **kwargs) -> Union[List[CkanActionResponse], list]:
        """
        Multiply request with a limited length until no more data is transmitted thanks to the offset parameter.
        List implementation returns the list of the unitary function return values.

        :param api_fun: function to call, typically a unitary request function
        :param params: api_fun must accept params argument in order to transmit other values and enforce the offset parameter
        :param limit: api_fun must accept limit argument in order to update the limit value
        :param offset: api_fun must accept offset argument in order to update the offset value
        :param search_all: if False, only the first request is operated
        :param kwargs: additional keyword arguments to pass to api_fun
        :return:
        """
        return list(self._request_all_results_generator(api_fun=api_fun, params=params, limit=limit, offset=offset,
                                                            search_all=search_all, **kwargs))

    def is_url_internal(self, url:str) -> bool:
        """
        Tests whether a url points to the same server as the CKAN url.

        :param url:
        :return:
        """
        # TODO: improve the url matching test
        return url.startswith(self.url)

    def test_ckan_url_reachable(self, raise_error:bool=False) -> bool:
        """
        Test if the CKAN URL is reachable with a HEAD request.
        This does not check it is really a CKAN server and does not check authentication.
        """
        error_message = self.download_url_proxy_test_head(self.url, raise_error=raise_error, context="CKAN URL test")
        return error_message is None




#!python3
# -*- coding: utf-8 -*-
"""
The basic file format for DataStore: CSV
"""
from typing import Union, Dict
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords
from ckanapi_harvesters.auxiliary.ckan_auxiliary import df_download_to_csv_kwargs
from ckanapi_harvesters.harvesters.file_formats.file_format_abc import FileFormatABC


csv_file_upload_read_csv_kwargs = dict(dtype=str, keep_default_na=False)


class CsvFileFormat(FileFormatABC):
    def __init__(self, read_csv_kwargs: dict=None, to_csv_kwargs: dict=None) -> None:
        if read_csv_kwargs is None: read_csv_kwargs = csv_file_upload_read_csv_kwargs
        if to_csv_kwargs is None: to_csv_kwargs = df_download_to_csv_kwargs
        self.read_csv_kwargs:dict = read_csv_kwargs
        self.to_csv_kwargs:dict = to_csv_kwargs

    def read_file(self, file_path: str, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        return pd.read_csv(file_path, **self.read_csv_kwargs)

    def read_buffer(self, buffer: io.StringIO, fields: Union[Dict[str, CkanField],None]) -> Union[pd.DataFrame, ListRecords]:
        return pd.read_csv(buffer, **self.read_csv_kwargs)

    def write_file(self, df: pd.DataFrame, file_path: str, fields: Union[Dict[str, CkanField],None]) -> None:
        df.to_csv(file_path, index=False, **self.to_csv_kwargs)

    def write_in_memory(self, df: pd.DataFrame, fields: Union[Dict[str, CkanField],None]) -> bytes:
        buffer = io.StringIO()
        df.to_csv(buffer, index=False, **self.to_csv_kwargs)
        return buffer.getvalue().encode("utf8")

    def copy(self):
        return CsvFileFormat(self.read_csv_kwargs, self.to_csv_kwargs)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to convert formats between database and local files.
"""
from typing import Dict, List, Callable, Any, Tuple, Union, Set
import copy

import pandas as pd

from ckanapi_harvesters.builder.builder_errors import MissingCodeFileError
from ckanapi_harvesters.auxiliary.external_code_import import PythonUserCode
from ckanapi_harvesters.auxiliary.list_records import ListRecords


def simple_upload_fun(df_local: pd.DataFrame) -> None:
    for field in df_local.columns:
        if df_local[field].dtype == pd.Timestamp:
            df_local[field] = df_local[field].apply(pd.Timestamp.isoformat)  # ISO-8601 format


class DataSchemeConversion:
    def __init__(self, *, df_upload_fun:Callable[[pd.DataFrame], Any] = None,
                 df_download_fun:Callable[[pd.DataFrame], Any] = None):
        """
        Class to convert between local data formats and database formats

        :param df_upload_fun:
        :param df_download_fun:
        """
        self.df_upload_fun:Union[Callable[[Any, Any], Union[ListRecords, pd.DataFrame]], None] = df_upload_fun
        self.df_download_fun:Union[Callable[[pd.DataFrame, Any], pd.DataFrame], None] = df_download_fun

    def copy(self):
        return copy.deepcopy(self)

    def df_upload_alter(self, df_local: Union[pd.DataFrame, Any], file_name:str=None, mapper_kwargs:dict=None, **kwargs) -> pd.DataFrame:
        """
        Apply used-defined df_upload_fun if present

        :param df_local: the dataframe to upload
        :return: the dataframe ready for upload, converted in the format of the database
        """
        if mapper_kwargs is None: mapper_kwargs = {}
        mapper_kwargs["file_name"] = file_name
        if file_name is not None and isinstance(df_local, pd.DataFrame):
            df_local.attrs["source"] = file_name
        df_database = df_local
        if self.df_upload_fun is not None:
            # df_database = df_local.copy()  # unnecessary copy
            df_upload_fun = self.df_upload_fun
            df_database = df_upload_fun(df_database, **mapper_kwargs, **kwargs)
        if not isinstance(df_database, pd.DataFrame):
            if isinstance(df_database, ListRecords):
                pass  # also accept ListRecords (List[dict])
            elif self.df_upload_fun is None:
                raise TypeError("No upload function was defined to convert the data format to a DataFrame")
            else:
                raise TypeError("df_upload_fun must return a DataFrame")
        return df_database

    def df_download_alter(self, df_database:pd.DataFrame, file_query:dict=None, mapper_kwargs:dict=None, **kwargs) -> pd.DataFrame:
        """
        Apply used-defined df_download_fun if present.
        df_download_fun should be the reverse function of df_upload_fun

        :param df_database: the downloaded dataframe from the database
        :return: the dataframe ready to save, converted in the local format
        """
        if mapper_kwargs is None: mapper_kwargs = {}
        mapper_kwargs["file_query"] = file_query
        if file_query is not None:
            df_database.attrs["query"] = file_query
        df_local = df_database
        if self.df_download_fun is not None:
            # df_local = df_database.copy()  # unnecessary copy
            df_download_fun = self.df_download_fun
            df_local = df_download_fun(df_local, **mapper_kwargs, **kwargs)
        return df_local

    def _connect_aux_functions(self, module: PythonUserCode, aux_upload_fun_name:str, aux_download_fun_name:str) -> None:
        if (aux_upload_fun_name or aux_download_fun_name) and module is None:
            raise MissingCodeFileError()
        if aux_upload_fun_name:
            self.df_upload_fun = module.function_pointer(aux_upload_fun_name)
        if aux_download_fun_name:
            self.df_download_fun = module.function_pointer(aux_download_fun_name)

    def get_necessary_fields(self) -> Set[str]:
        return set()




#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Union, Tuple
import time
from warnings import warn
import io

import pandas as pd

from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice, CkanState
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_action import CkanActionResponse
# from ckanapi_harvesters.auxiliary.list_records import records_to_df
from ckanapi_harvesters.auxiliary.ckan_auxiliary import upload_prepare_requests_files_arg, RequestType, json_encode_params
from ckanapi_harvesters.auxiliary.ckan_errors import (ReadOnlyError, IntegrityError, MaxRequestsCountError,
                                                      UnexpectedError, InvalidParameterError, DataStoreNotFoundError)
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy

from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.ckan_api.ckan_api_3_policy import CkanApiPolicyParams
from ckanapi_harvesters.ckan_api.ckan_api_3_policy import CkanApiPolicy

from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.auxiliary.ckan_auxiliary import df_upload_to_csv_kwargs




class CkanApiReadWriteParams(CkanApiPolicyParams):
    # not read-only by default
    default_readonly:bool = False

    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.default_limit_write: Union[int,None] = self.default_limit_read  # limit the number of entries per upsert (used as default value)
        self.default_force: bool = True  # set to True to edit a read-only resource
        self.read_only: bool = self.default_readonly
        self.submit_delay: float = 2.0      # delay between requests when running datapusher_submit
        self.submit_timeout: float = 90.0   # maximum wait time after datapusher_submit

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiReadWriteParams()
        super().copy(dest=dest)
        dest.default_limit_write = self.default_limit_write
        dest.default_force = self.default_force
        dest.read_only = self.read_only
        dest.submit_delay = self.submit_delay
        dest.submit_timeout = self.submit_timeout
        return dest


class CkanApiReadWrite(CkanApiPolicy):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements requests to write data to the CKAN server resources / DataStores.
    """

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org: str = None, params:CkanApiPolicyParams=None,
                 map:CkanMap=None, policy: CkanPackageDataFormatPolicy = None, policy_file:str=None,
                 data_cleaner_upload:CkanDataCleanerABC=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param policy: data format policy to use with policy_check function
        :param policy_file: path to a JSON file containing the data format policy to use with policy_check function
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param policy: data format policy to be used with the policy_check function.
        :param policy_file: path to a JSON file containing the data format policy to load.
        :param data_cleaner_upload: data cleaner object to use before uploading to a CKAN DataStore.
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, map=map, policy=policy, policy_file=policy_file, identifier=identifier)
        self.data_cleaner_upload: Union[CkanDataCleanerABC, None] = data_cleaner_upload
        if params is None:
            params = CkanApiReadWriteParams()
        if proxies is not None:
            params.proxies = proxies
        self.params: CkanApiReadWriteParams = params

    def copy(self, new_identifier: str = None, *, dest=None):
        if dest is None:
            dest = CkanApiReadWrite()
        super().copy(new_identifier=new_identifier, dest=dest)
        dest.data_cleaner_upload = self.data_cleaner_upload.copy() if self.data_cleaner_upload is not None else None
        return dest

    def full_unlock(self, unlock:bool=True,
                    *, no_ca:bool=None, external_url_resource_download:bool=None) -> None:
        """
        Function to unlock full capabilities of the CKAN API

        :param unlock:
        :return:
        """
        super().full_unlock(unlock, no_ca=no_ca, external_url_resource_download=external_url_resource_download)
        self.params.default_force = unlock
        self.params.read_only = not unlock

    def set_limits(self, limit_read:Union[int,None], limit_write:int=None) -> None:
        """
        Set default query limits. If only one argument is provided, it applies to both limits.

        :param limit_read: default limit for read requests
        :param limit_write: default limit for upsert (write) requests
        :return:
        """
        super().set_limits(limit_read)
        if limit_write is not None:
            self.params.default_limit_write = limit_write
        else:
            self.params.default_limit_write = limit_read

    def set_submit_timeout(self, submit_timeout:float, submit_delay:float=None) -> None:
        """
        Set timeout for the datastore_wait method. This is called after datastore_submit.

        :param submit_timeout: timeout after which a TimeoutError is raised
        :param submit_delay: delay between requests to peer on DataStore initialization
        :return:
        """
        self.params.submit_timeout = submit_timeout
        if submit_delay is not None:
            self.params.submit_delay = submit_delay

    ## DataStore insertions ------------------
    def _api_datastore_upsert_raw(self, records:Union[dict, List[dict], pd.DataFrame], resource_id:str, *,
                                  method:Union[UpsertChoice,str], params:dict=None, force:bool=None, dry_run:bool=False,
                                  last_insertion:bool=True) -> CkanActionResponse:
        """
        API call to api_datastore_upsert.

        :param records: records, preferably in a pandas DataFrame - they will be converted to a list of dictionaries.
        :param resource_id: destination resource id
        :param method: see UpsertChoice (insert, update or upsert)
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param params: additional parameters
        :param dry_run: set to True to abort transaction instead of committing, e.g. to check for validation or type errors
        :param last_insertion: trigger for calculate_record_count
        (doc: updates the stored count of records, used to optimize datastore_search in combination with the
        total_estimation_threshold parameter. If doing a series of requests to change a resource, you only need to set
        this to True on the last request.)
        :return: the inserted records as a pandas DataFrame, from the server response
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        if force is None: force = self.params.default_force
        if method is not None: method = str(method)
        params["resource_id"] = resource_id
        params["force"] = force
        params["dry_run"] = dry_run
        #  doc calculate_record_count: updates the stored count of records, used to optimize datastore_search in combination with the
        # total_estimation_threshold parameter. If doing a series of requests to change a resource, you only need to set
        # this to True on the last request.
        params["calculate_record_count"] = last_insertion
        # params["force_indexing"] = last_insertion
        N = self.map.get_datastore_len(resource_id, error_not_mapped=False)
        has_datastore_info = N is not None
        format = None  # API does not support csv or other formats
        mode_df = True
        if records is not None:
            method = method.lower()
            params["method"] = method
            if isinstance(records, dict):
                records = pd.DataFrame.from_dict(records)
            elif isinstance(records, list):
                # records = records_to_df(records)
                mode_df = False
            else:
                assert(isinstance(records, pd.DataFrame))
            n_upsert = len(records)
            if not mode_df:
                params["records"] = records
                format = None
            elif format is None or format == "objects":
                params["records"] = records.to_dict(orient='records')
            else:
                # dead code
                fields_id_list = records.columns.tolist()
                params["fields"] = fields_id_list  # [{"id": id} for id in fields_id_list]
                params["records_format"] = format
                if format == "csv":
                    params["records"] = records.to_csv(index=False, header=False, **df_upload_to_csv_kwargs)
                elif format == "lists":
                    params["records"] = records.values.tolist()
                else:
                    raise NotImplementedError()
        else:
            # possibility to call with None records and method to trigger row counts etc.
            # this request may not be useful
            assert(method is None)
            n_upsert = 0
        # json encode here in the case there are NaN values, not supported by the requests encoder
        data_payload, json_headers = json_encode_params(params)
        response = self._api_action_request(f"datastore_upsert", method=RequestType.Post,
                                            data=data_payload, headers=json_headers)
        # response = self._api_action_request(f"datastore_upsert", method=RequestType.Post,
        #                                     json=params)
        if response.success:
            if method is not None:
                n_return = len(response.result["records"])
                if has_datastore_info and not dry_run and method == "insert":
                    # in modes other than insert could be updated rather than inserted
                    self.map._update_datastore_len(resource_id, N + n_return)
                assert_or_raise(n_return == n_upsert, IntegrityError("Returned dataframe does not match number of requested rows"))
            return response
        else:
            raise response.default_error(self)

    def _api_datastore_upsert(self, records:Union[dict, List[dict], pd.DataFrame], resource_id:str, *,
                              method:Union[UpsertChoice,str], params:dict=None, force:bool=None, dry_run:bool=False,
                              last_insertion:bool=True, return_df:bool=None) -> Union[pd.DataFrame, List[dict], dict]:
        mode_df = True
        if isinstance(records, list):
            mode_df = False
        if return_df is None:
            return_df = mode_df
        response = self._api_datastore_upsert_raw(records=records, resource_id=resource_id, method=method,
                                                  params=params, force=force, dry_run=dry_run,
                                                  last_insertion=last_insertion)
        if method is not None:
            if return_df:
                response_df = pd.DataFrame.from_dict(response.result["records"])
                self._rx_records_df_clean(response_df)
                return response_df
            else:
                return response.result["records"]
        else:
            return response.result

    def datastore_upsert_last_line(self, resource_id:str):
        """
        Apply last line treatments to a resource.
        """
        return self._api_datastore_upsert(None, resource_id=resource_id, method=None, last_insertion=True)

    def datastore_upsert(self, records:Union[dict, List[dict], pd.DataFrame], resource_id:str, *,
                         dry_run:bool=False, limit:int=None, offset:int=0, force:bool=None,
                         method:Union[UpsertChoice,str]=UpsertChoice.Upsert, apply_last_condition:bool=True,
                         always_last_condition:bool=None, return_df:bool=None,
                         data_cleaner:CkanDataCleanerABC=None, params:dict=None) -> Union[pd.DataFrame, List[dict]]:
        """
        Encapsulation of _api_datastore_upsert to cut the requests to a limited number of rows.

        :see: _api_datastore_upsert()
        :param records: records, preferably in a pandas DataFrame - they will be converted to a list of dictionaries.
        :param resource_id: destination resource id
        :param method: by default, set to Upsert
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param limit: number of records per transaction
        :param offset: number of records to skip - use to restart the transfer
        :param params: additional parameters
        :param dry_run: set to True to abort transaction instead of committing, e.g. to check for validation or type errors
        :param apply_last_condition: if True, the last upsert request applies the last insert operations (calculate_record_count and force_indexing).
        :param always_last_condition: if True, each request applies the last insert operations - default is False
        :return: the inserted records as a pandas DataFrame, from the server response
        """
        method_str = str(method)
        if apply_last_condition is None:
            apply_last_condition = True
        if always_last_condition is None:
            always_last_condition = False
        mode_df = True
        assert(records is not None)
        if isinstance(records, dict):
            records = pd.DataFrame.from_dict(records)
        elif isinstance(records, list):
            # records = records_to_df(records)
            mode_df = False
        else:
            assert(isinstance(records, pd.DataFrame))
        if data_cleaner is None:
            data_cleaner = self.data_cleaner_upload
        if data_cleaner is not None:
            datastore_info = self.get_datastore_info_or_request_of_id(resource_id=resource_id, error_not_found=True)
            records = data_cleaner.clean_records(records, known_fields=datastore_info.fields_dict, inplace=True)
            data_cleaner.apply_new_fields_request(self, resource_id=resource_id)
        if return_df is None:
            return_df = mode_df
        if limit is None: limit = self.params.default_limit_write
        if limit is None:
            # direct API call with one request
            if self.params.store_last_response_debug_info:
                self.debug.multi_requests_last_successful_offset = offset
            return self._api_datastore_upsert(records, return_df=return_df,
                                              method=method_str, dry_run=dry_run, resource_id=resource_id,
                                              force=force, params=params,
                                              last_insertion=apply_last_condition or always_last_condition)
        assert_or_raise(limit > 0, InvalidParameterError("limit"))
        n = len(records)
        if self.params.store_last_response_debug_info:
            self.debug.multi_requests_last_successful_offset = offset
        requests_count = 0
        last_insertion = True
        df, returned_rows = None, None
        if return_df:
            df = None
        else:
            returned_rows = []
        start = time.time()
        current = start
        timeout = False
        n_cum = 0
        while offset < n and requests_count < self.params.max_requests_count and not timeout:
            last_insertion = offset+limit >= n
            i_end_add = min(n, offset+limit)
            n_add = i_end_add-1 - offset + 1
            if self.params.verbose_multi_requests:
                print(f"{self.identifier} Multi-requests upsert {requests_count} to add {n_add} records ...")
            if mode_df:
                df_upsert = records.iloc[offset:i_end_add]
            else:
                df_upsert = records[offset:i_end_add]
            df_add = self._api_datastore_upsert(df_upsert, return_df=return_df,
                                                method=method_str, dry_run=dry_run, resource_id=resource_id,
                                                force=force, params=params,
                                                last_insertion=(last_insertion and apply_last_condition) or always_last_condition)
            n_cum += len(df_add)
            assert_or_raise(len(df_add) == n_add, IntegrityError("Second check on response len failed in datastore_upsert"))  # consistency check, in double of _api_datastore_upsert
            if self.params.store_last_response_debug_info:
                self.debug.multi_requests_last_successful_offset = offset
            if return_df:
                if df is None:
                    # 1st execution: pandas cannot concatenate with an empty DataFrame => use None as indicator
                    assert(df_add is not None)
                    df = df_add
                else:
                    df = pd.concat([df, df_add], ignore_index=True)
            else:
                returned_rows = returned_rows + df_add
            if self.params.multi_requests_time_between_requests > 0 and not last_insertion:
                time.sleep(self.params.multi_requests_time_between_requests)
            if not last_insertion:
                assert_or_raise(n_add == limit, IntegrityError("datastore_upsert implementation is wrong"))
            offset += limit
            requests_count += 1
            current = time.time()
            timeout = current - start > self.params.multi_requests_timeout
        if return_df:
            if df is None:
                df = pd.DataFrame()  # always return a DataFrame object and not None
            df.attrs["requests_count"] = requests_count
            df.attrs["elapsed_time"] = current - start
            df.attrs["offset"] = offset
        if self.params.verbose_multi_requests:
            print(f"{self.identifier} Multi-requests upsert done to add {n_cum} records done in {requests_count} requests and {round(current - start, 2)} seconds.")
        if timeout:
            raise TimeoutError()
        if requests_count >= self.params.max_requests_count:
            raise MaxRequestsCountError()
        assert_or_raise(last_insertion, UnexpectedError("last_insertion should be True at last iteration"))
        if mode_df:
            return df
        else:
            return returned_rows

    def datastore_insert(self, records:Union[dict, List[dict], pd.DataFrame], resource_id:str, *,
                         dry_run:bool=False, limit:int=None, offset:int=0, apply_last_condition:bool=True,
                         always_last_condition:bool=None,
                         data_cleaner:CkanDataCleanerABC=None, force:bool=None, params:dict=None) -> pd.DataFrame:
        """
        Alias function to insert data in a DataStore using datastore_upsert.

        :see: _api_datastore_upsert()
        :param records: records, preferably in a pandas DataFrame - they will be converted to a list of dictionaries.
        :param resource_id: destination resource id
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param params: additional parameters
        :param dry_run: set to True to abort transaction instead of committing, e.g. to check for validation or type errors
        :return: the inserted records as a pandas DataFrame, from the server response
        """
        return self.datastore_upsert(records, resource_id, dry_run=dry_run, limit=limit, offset=offset,
                                     method=UpsertChoice.Insert, apply_last_condition=apply_last_condition,
                                     always_last_condition=always_last_condition, data_cleaner=data_cleaner,
                                     force=force, params=params)

    def datastore_update(self, records:Union[dict, List[dict], pd.DataFrame], resource_id:str, *,
                         dry_run:bool=False, limit:int=None, offset:int=0, apply_last_condition:bool=True,
                         always_last_condition:bool=None,
                         data_cleaner:CkanDataCleanerABC=None, force:bool=None, params:dict=None) -> pd.DataFrame:
        """
        Alias function to update data in a DataStore using datastore_upsert.
        The update is performed based on the DataStore primary keys

        :see: _api_datastore_upsert()
        :param records: records, preferably in a pandas DataFrame - they will be converted to a list of dictionaries.
        :param resource_id: destination resource id
        :param force: set to True to edit a read-only resource. If not provided, this is overridden by self.default_force
        :param params: additional parameters
        :param dry_run: set to True to abort transaction instead of committing, e.g. to check for validation or type errors
        :return: the inserted records as a pandas DataFrame, from the server response
        """
        return self.datastore_upsert(records, resource_id, dry_run=dry_run, limit=limit, offset=offset,
                                     method=UpsertChoice.Update, apply_last_condition=apply_last_condition,
                                     always_last_condition=always_last_condition, data_cleaner=data_cleaner,
                                     force=force, params=params)


    ## Resource updates ------------------
    def _api_resource_patch(self, resource_id:str, *, name:str=None, format:str=None, description:str=None, title:str=None,
                            state:CkanState=None,
                            df:pd.DataFrame=None, file_path:str=None, url:str=None, files=None,
                            payload: Union[bytes, io.BufferedIOBase] = None, payload_name: str = None,
                            params:dict=None) -> CkanResourceInfo:
        """
        Call to resource_patch API. This call can be used to change the resource parameters via params (cf. API documentation)
        or to reupload the resource file into FileStore.
        The latter action replaces the current resource. If it is a DataStore, it is reset to the new contents of the file.
        The file can be transmitted either as an url, a file path or a pandas DataFrame.
        The files argument can pass through these arguments to the requests.post function.
        A call to datapusher_submit() could be required to take immediately into account the newly downloaded file.

        :see: _api_resource_create
        :see: resource_create
        :param resource_id: resource id
        :param url: url of the resource to replace resource
        :param params: parameters such as name, format, resource_type can be changed

        For file uploads, the following parameters are taken, by order of priority:
        See upload_prepare_requests_files_arg for an example of formatting.

        :param files: files pass through argument to the requests.post function. Use to send other data formats.
        :param payload: bytes to upload as a file
        :param payload_name: name of the payload to use (associated with the payload argument) - this determines the format recognized in CKAN viewers.
        :param file_path: path of the file to transmit (binary and text files are supported here)
        :param df: pandas DataFrame to replace resource

        :return:
        """
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        params["id"] = resource_id
        if description is not None:
            params["description"] = description
        if title is not None:
            params["title"] = title
        if name is not None:
            params["name"] = name
        if format is not None:
            params["format"] = format
        if state is not None:
            params["state"] = str(state)
        files = upload_prepare_requests_files_arg(files=files, file_path=file_path, df=df, payload=payload, payload_name=payload_name)
        if url is not None:
            params["url"] = url
            params["clear_upload"] = True
            assert(files is None)
        if files is not None:
            response = self._api_action_request(f"resource_patch", method=RequestType.Post,
                                                files=files, data=params)
        else:
            response = self._api_action_request(f"resource_patch", method=RequestType.Post, json=params)
        if response.success:
            resource_info = CkanResourceInfo(response.result)
            self.map._record_resource_update(resource_info)
            return resource_info
        else:
            raise response.default_error(self)

    def resource_patch(self, resource_id:str, *, name:str=None, format:str=None, description:str=None, title:str=None,
                       state:CkanState=None,
                       df:pd.DataFrame=None, file_path:str=None, url:str=None, files=None,
                       payload: Union[bytes, io.BufferedIOBase] = None, payload_name: str = None,
                       params:dict=None) -> CkanResourceInfo:
        # function alias
        return self._api_resource_patch(resource_id, name=name, format=format, description=description, state=state,
                                        title=title, df=df, file_path=file_path, url=url, files=files,
                                        payload=payload, payload_name=payload_name, params=params)

    ### DataPusher submit ------------------
    def _api_datapusher_submit(self, resource_id: str, *, params: dict = None) -> bool:
        """
        Call to API action datapusher_submit. This triggers the normally asynchronous DataPusher service for a given resource.

        :param resource_id: resource id
        :param params:
        :return:
        """
        if params is None: params = {}
        params["resource_id"] = resource_id
        response = self._api_action_request(f"datapusher_submit", method=RequestType.Post, json=params)
        if response.success:
            return response.result
        else:
            raise response.default_error(self)

    def datastore_wait(self, resource_id: str, *,
                       apply_delay:bool=True, error_timeout:bool=True) -> Tuple[int, float]:
        """
        Wait until a DataStore has at least one row.
        The delay between requests to peer on the presence of the DataStore is given by the class attribute submit_delay.
        If the loop exceeds submit_timeout, an exception is raised.

        :param resource_id:
        :param apply_delay:
        :param error_timeout: option to raise an exception in case of timeout
        :return:
        """
        if self.params.submit_delay <= 0 or not apply_delay:
            return 0, 0.0
        # resource_info = self.resource_show(resource_id)
        # init_timestamp = resource_info.last_modified
        # current_timestamp = init_timestamp
        if self.params.verbose_request:
            print(f"Waiting for data treatments on DataStore {resource_id}...")
        start = time.time()
        current = start
        timeout = False
        counter = 0
        df = pd.DataFrame()  # empty DataFrame
        while not timeout and df.empty:  # current_timestamp <= init_timestamp:
            time.sleep(self.params.submit_delay)
            # resource_info = self.resource_show(resource_id)
            # current_timestamp = resource_info.last_modified
            try:
                df = self.datastore_search(resource_id, limit=1, search_all=False, search_method=True)
            except DataStoreNotFoundError:
                pass
            current = time.time()
            timeout = (current - start) > self.params.submit_timeout
            counter += 1
        if timeout:
            if error_timeout:
                raise TimeoutError("datastore_wait")
            else:
                msg = str(TimeoutError("datastore_wait"))
                warn(msg)
        if self.params.verbose_request:
            print(f"Resource updated after {current - start} seconds ({counter} iterations)")
        return counter, current - start

    def datastore_submit(self, resource_id: str,
                          *, apply_delay:bool=True, error_timeout:bool=True,
                          params: dict = None) -> bool:
        """
        Submit file to re-initiate DataStore, using the preferred method.
        Current method is datapusher_submit.
        This encapsulation includes a call to datastore_wait.

        :param resource_id:
        :param apply_delay: Keep true to wait until the datastore is ready (a datastore_search query is performed as a test)
        :param params:
        :return:
        """
        result = self._api_datapusher_submit(resource_id, params=params)
        self.datastore_wait(resource_id, apply_delay=apply_delay, error_timeout=error_timeout)
        return result

    # def datapusher_submit_insert(self, resource_id: str, *, params: dict = None) -> dict:
    #     # idea: modify datapusher such as it would upsert data instead of replacing the entire datastore
    #     raise NotImplementedError()
    #     if params is None: params = {}
    #     params["insert"] = True
    #     return self._api_datapusher_submit(resource_id, params)











#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing for lists of values such as tags
"""
from typing import List, Any, Iterable, Union, Dict, Set
from abc import ABC, abstractmethod
from warnings import warn
import re

from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import ListChoiceMode
from ckanapi_harvesters.policies.data_format_policy_defs import StringValueSpecification
from ckanapi_harvesters.policies.data_format_policy_abc import DataPolicyElementABC


extra_group_name = "extra"


class ValueListPolicy(DataPolicyElementABC):
    _group_type_str = "group"

    def __init__(self, list_specs:List[StringValueSpecification]=None, group_name:str=None,
                 value_select:ListChoiceMode=ListChoiceMode.Any,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if list_specs is None:
            list_specs = []
        self.list_specs:List[StringValueSpecification] = list_specs
        self.value_select: ListChoiceMode = value_select
        self.group_name: str = group_name

    def to_dict(self) -> dict:
        d = {}
        if self.group_name:
            d["group_name"] = self.group_name
        d.update(super().to_dict())
        d.update({"values": [spec.to_dict() for spec in self.list_specs],
                  "value_select": self.value_select.name})
        return d

    def list_specs_str(self) -> List[str]:
        return [value_spec.value for value_spec in self.list_specs]

    @staticmethod
    def from_dict(d:dict) -> "ValueListPolicy":
        obj = ValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.group_name = d["group_name"] if len(d["group_name"]) > 0 else None
        self.list_specs = [StringValueSpecification.from_dict(value)  for value in d["values"]]
        self.value_select = ListChoiceMode.from_str(d["value_select"]) if "value_select" in d.keys() else ListChoiceMode.Any

    def enforce(self, values: Union[str, List[str]], *, context:str=None,
                verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        if self.group_name is not None:
            context = context + " / " + self._group_type_str + " " + self.group_name
        success = True
        spec = [tag_spec.value for tag_spec in self.list_specs]
        if values is None:
            values = []
        elif isinstance(values, str):
            values = values.split(ckan_tags_sep)
        values = list(set(values).intersection(set(spec)))
        msg = None
        value_context = context + " / value '" + ','.join(values).join(values) + "'"
        if (self.value_select == ListChoiceMode.MaxOne and len(values) > 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. Max one value is admitted within {spec}.")
        if (self.value_select == ListChoiceMode.NoExtra and len(values) > 0):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. No values can be selected for this group ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryOne and not len(values) == 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Exactly one value must be present for value list group '{self.group_name}' ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryMulti and not len(values) < 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"At least one value must be present for value list group '{self.group_name}' ({spec}).")
        if not success:
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
        if len(spec) > 0:
            for tag in values:
                success &= self._enforce_unit_string(tag, spec, context=context, verbose=verbose, buffer=buffer)
        return success


class ExtraValueListPolicy(ValueListPolicy):
    def __init__(self, list_specs:List[StringValueSpecification]=None,
                 value_select:ListChoiceMode=ListChoiceMode.Any,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(list_specs=list_specs, group_name=extra_group_name,
                         value_select=value_select, mandatory=mandatory, error_level=error_level)

    @staticmethod
    def from_ValueListPolicy(value: ValueListPolicy) -> "ExtraValueListPolicy":
        obj = ExtraValueListPolicy()
        obj.__dict__.update(value.__dict__)
        return obj

    @staticmethod
    def from_dict(d:dict) -> "ExtraValueListPolicy":
        obj = ExtraValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def enforce(self, values: Union[str, List[str]], *, context:str=None,
                verbose: bool = True, buffer:List[DataPolicyError]=None, extra_spec_rm:Set[str]=None) -> bool:
        if self.group_name is not None:
            context = context + " / group " + self.group_name
        success = True
        spec = [tag_spec.value for tag_spec in self.list_specs]
        if values is None:
            values = []
        elif isinstance(values, str):
            values = values.split(ckan_tags_sep)
        values = list(set(values) - extra_spec_rm)
        msg = None
        value_context = context + " / value '" + ','.join(values).join(values) + "'"
        if (self.value_select == ListChoiceMode.MaxOne and len(values) > 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. Max one value is admitted within {spec}.")
        if (self.value_select == ListChoiceMode.NoExtra and len(values) > 0):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Too many values for value list group '{self.group_name}'. No values can be selected for this group ({spec}). Admitted values from other groups: {extra_spec_rm}")
        if (self.value_select == ListChoiceMode.MandatoryOne and not len(values) == 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"Exactly one value must be present for value list group '{self.group_name}' ({spec}).")
        if (self.value_select == ListChoiceMode.MandatoryMulti and not len(values) < 1):
            success = False
            msg = DataPolicyError(value_context, self.error_level, f"At least one value must be present for value list group '{self.group_name}' ({spec}).")
        if not success:
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
        if len(spec) > 0:
            for tag in values:
                success &= self._enforce_unit_string(tag, spec, context=context, verbose=verbose, buffer=buffer)
        return success


class GroupedValueListPolicy(DataPolicyElementABC):
    def __init__(self, value_group_specs:List[ValueListPolicy]=None,
                 extra_values:ExtraValueListPolicy=None,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if value_group_specs is None:
            value_group_specs = []
        self.value_group_specs:List[ValueListPolicy] = value_group_specs
        self.extra_values_spec:ExtraValueListPolicy = extra_values
        self._extract_extra_values()

    def _extract_extra_values(self):
        i_rm = []
        extra_values = self.extra_values_spec
        for i, value_group_spec in enumerate(self.value_group_specs):
            if value_group_spec.group_name.lower() == extra_group_name.lower():
                assert(extra_values is None)
                extra_values = ExtraValueListPolicy.from_ValueListPolicy(value_group_spec)
                i_rm.append(i)
        for i in reversed(i_rm):
            self.value_group_specs.pop(i)
        if extra_values is not None:
            self.extra_values_spec:ExtraValueListPolicy = extra_values

    def to_dict(self) -> dict:
        d = super().to_dict()
        if self.extra_values_spec is not None:
            self.extra_values_spec.group_name = extra_group_name
            extra_values_dict = [self.extra_values_spec.to_dict()]
        else:
            extra_values_dict = []
        d.update({"groups": [spec.to_dict() for spec in self.value_group_specs] + extra_values_dict})
        return d

    @staticmethod
    def from_dict(d:dict) -> "GroupedValueListPolicy":
        obj = GroupedValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict, child_cls:type=None):
        super()._load_from_dict(d)
        if child_cls is None:
            child_cls = ValueListPolicy
        self.value_group_specs = [child_cls.from_dict(group_spec) for group_spec in d["groups"]]
        self.extra_values_spec = None
        self._extract_extra_values()

    def enforce(self, values: Union[str, List[str]], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = True
        extra_spec_rm = set()
        for value_group_spec in self.value_group_specs:
            if not value_group_spec.group_name == extra_group_name.lower():
                extra_spec_rm = extra_spec_rm.union({tag_spec.value for tag_spec in value_group_spec.list_specs})
        for value_group_spec in self.value_group_specs:
            success &= value_group_spec.enforce(values, context=context, verbose=verbose, buffer=buffer)
        if self.extra_values_spec is not None:
            self.extra_values_spec.group_name = extra_group_name
            success &= self.extra_values_spec.enforce(values, context=context, verbose=verbose, buffer=buffer, extra_spec_rm=extra_spec_rm)
        return success


class SingleValueListPolicy(DataPolicyElementABC):
    def __init__(self, base_list:ValueListPolicy=None, extra_values:ListChoiceMode=ListChoiceMode.Any, mandatory:bool=False):
        super().__init__(mandatory=mandatory)
        self.base_list: GroupedValueListPolicy = GroupedValueListPolicy(extra_values=ExtraValueListPolicy(value_select=extra_values))
        self.update_base_list(base_list)

    def to_dict(self) -> dict:
        return self.base_list.to_dict()

    @staticmethod
    def from_dict(d:dict) -> "SingleValueListPolicy":
        obj = SingleValueListPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        # super()._load_from_dict(d)
        self.base_list._load_from_dict(d)

    def update_base_list(self, base_list:ValueListPolicy):
        if base_list is not None:
            base_list.group_name = "base"
            self.base_list.value_group_specs = [base_list]
        else:
            self.base_list.value_group_specs = []

    def enforce(self, values: Union[str, List[str]], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = self.base_list.enforce(values, context=context, verbose=verbose, buffer=buffer)
        return success
